/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 *
 * This code is a derivative work of other copyrighted software.  See the
 * file COPYRIGHTS for more information on terms and conditions.
 */

extern "C" {
#include <stk.h>
}
#include <narray.hh>

extern "C" {
  int internal_wrap_filter(float *, int, int, float *, int, int,
			   int, int, int, int, float *, int);
  int internal_wrap_expand(float *, int, int, float *, int, int,
			   int, int, int, int, float *, int);
  int internal_filter(float *, int, int, float *, float *, int, int,
		      int, int, int, int, float *, char *);
  int internal_expand(float *, int, int, float *, float *, int, int,
		      int, int, int, int, float *, char *);
}

static void
my_err(const char *function, const char *explanation, const SCM object)
{
  char error_message[256];
  uint l;

  strncpy(error_message, function, 256);
  l = 254 - strlen(error_message);
  strcat(error_message, ": ");
  strncat(error_message, explanation, l);

  STk_err(error_message, object);
}

static void
convolve_and_downsample(float *ddata, int dx, int dy,
			float *kdata, int kx, int ky,
			float *rdata,
			int xdelay, int xrate, int ydelay, int yrate,
			char *edge_type)
{
  if(strcmp(edge_type, "periodic")==0)
    internal_wrap_filter(ddata, dx, dy,
			 kdata, kx, ky, 
			 xdelay, xrate, ydelay, yrate,
			 rdata, 0);
  else {
    float *tmp = (float *)calloc(kx * ky, sizeof(float));
    internal_filter(ddata, dx, dy,
		    kdata, tmp, kx, ky,
		    xdelay, xrate, ydelay, yrate,
		    rdata, edge_type);
  }
}  
  
static void
convolve_and_upsample(float *ddata, int dx, int dy,
		      float *kdata, int kx, int ky,
		      float *rdata,
		      int xdelay, int xrate, int ydelay, int yrate,
		      char *edge_type)
{
  if(strcmp(edge_type, "periodic")==0)
    internal_wrap_expand(ddata, dx, dy,
			 kdata, kx, ky, 
			 xdelay, xrate, ydelay, yrate,
			 rdata, 0);
  else {
    float *tmp = (float *)calloc(kx * ky, sizeof(float));
    internal_expand(ddata, dx, dy,
		    kdata, tmp, kx, ky,
		    xdelay, xrate, ydelay, yrate,
		    rdata, edge_type);
  }
}  

extern "C" PRIMITIVE
narray_convolve2(SCM arglist, int argcount)
{
  const char *function_name = "narray-convolve2";

  SCM args = STk_reverse(arglist);
  SCM dir, steps, delays, edges, kernel, data, result;
  SCM l;

  Real *kdata, *ddata, *rdata;
  index_type rshape[MAX_DIMENSION];

  bool ok;
  uint llength;
  uint kx, ky, dx, dy, rx, ry;
  Real xrate, yrate;
  uint xdelay, ydelay, dxoffset=0, dyoffset=0;
  char *edge_type;
  
  // setup defaults
  xrate = yrate = 1.0;
  xdelay = ydelay = 0;
  edge_type = "zero";
  dir = Intern("down");

  switch(argcount) {

  case 5:
    dir = CAR(args); args = CDR(args);
    if(!(SYMBOLP(dir) && (EQ(dir, Intern("up")) || EQ(dir, Intern("down")))))
      my_err(function_name,
	     "5th (optional) parameter, resampling direction not in '(up down)"
	     ,dir);

  case 4: // convolution with donwnsampling: arg gives list of rates in x, y

    steps = CAR(args); args = CDR(args);
    if(CONSP(steps))
      for(llength=0, ok=true, l=steps; l!=NIL; llength++, l=CDR(l))
	ok &= INTEGERP(CAR(l));
    if(!CONSP(steps) || llength!=2 || !ok)
      my_err(function_name,
	     "5th argument (step sizes) must be a length 2 list of rationals",
	     steps);
    else {
      xrate = numeric_value(CAR(steps));
      yrate = numeric_value(CAR(CDR(steps)));
    }
    
    
//   case 4: // convolution with delayed kernel: arg is list of delays in x, y
    
//     delays = CAR(args); args = CDR(args);
//     if(CONSP(delays))
//       for(llength=0, ok=true, l=delays; l!=NIL; llength++, l=CDR(l))
// 	ok &= INTEGERP(CAR(l));
//     if(!CONSP(delays) || llength!=2 || !ok)
//       my_err(function_name,
// 	     "4th argument (delays) must be a length 2 list of integers",
// 	     delays);
//     else {
//       if(INTEGER(CAR(delays)) > 0)
// 	xdelay = INTEGER(CAR(delays));
//       else
// 	dxoffset = -INTEGER(CAR(delays));
//       if(INTEGER(CAR(CDR(delays))) > 0)
// 	ydelay = INTEGER(CAR(CDR(delays)));
//       else
// 	dyoffset = -INTEGER(CAR(CDR(delays)));
//     }
    

  case 3: // convolution with special edge handling: list of symbols

    edges = CAR(args); args = CDR(args);
    if(SYMBOLP(edges)) {
      if(EQ(edges, Intern("periodic")))
	edge_type = "periodic";
      else if(EQ(edges, Intern("ws-symetric")))
	edge_type = "reflect1";
      else if(EQ(edges, Intern("hs-symetric")))
	edge_type = "reflect2";
      else if(EQ(edges, Intern("zero-padded")))
	edge_type = "zero";
      else
	my_err(function_name, 
	       "3rd argument (edges) must be a symbol from the set '(periodic ws-symetric zero-padded)",
	       edges);
    }
    else
      my_err(function_name, 
	     "3rd argument (edges) must be a symbol from the set '(periodic ws-symetric zero-padded)",
	     edges);
    
    
  case 2: // this is the convolution kernel: an narray
    
    kernel = CAR(args); args = CDR(args);
    if(!NARRAYP(kernel) || narray_instance(kernel)->dimension_ref() > 2)
      my_err(function_name,
	     "2nd argument (kernel) must be an narray with dimension <= 2",
	     kernel);
    else {
      kx = narray_instance(kernel)->shape_ref(0);
      if(narray_instance(kernel)->dimension_ref() > 1)
	ky = narray_instance(kernel)->shape_ref(1);
      else
	ky = 1;
      narray_instance(kernel)->convension_set(C);
      kdata = narray_instance(kernel)->storage_iref();
    }

    // Also get the data: an narray with rank >= the rank off the kernel
    data = CAR(args);
    if(!NARRAYP(data) || narray_instance(data)->dimension_ref() < narray_instance(kernel)->dimension_ref())
      my_err(function_name,
	     "1st argument (data) must be an narray with dimension <= 2",
	     data);
    else {
      dx = narray_instance(data)->shape_ref(0);
      if(narray_instance(data)->dimension_ref() > 1)
	dy = narray_instance(data)->shape_ref(1);
      else
	dy = 1;
      narray_instance(data)->convension_set(C);
      ddata = narray_instance(data)->storage_iref();
    }

    break;

  case 1:
  case 0:
    my_err(function_name, 
	   "USAGE: (narray-convolve2 image kernel edges '(xrate yrate) dir)", 
	   arglist);
  }


  /*
   * Hooray! we have all of our args and they all look correct... lets convolve
   */


  if(EQ(dir, Intern("down"))) {
    rx = (uint)(ceil((float)dx / xrate));
    ry = (uint)(ceil((float)dy / yrate));
    rdata = (Real*)calloc(rx*ry, sizeof(Real));
    convolve_and_downsample(ddata, dx, dy,
			    kdata, kx, ky,
			    rdata,
			    xdelay, (int)xrate, 
			    ydelay, (int)yrate,
			    edge_type);
  }
  else {
    rx = dx * xrate;
    ry = dy * yrate;
    rdata = (Real*)calloc(rx*ry, sizeof(Real));
    convolve_and_upsample(ddata, rx, ry,
			  kdata, kx, ky,
			  rdata,
			  xdelay, (int)(xrate), 
			  ydelay, (int)(yrate),
			  edge_type);
  }
  rshape[0] = rx; rshape[1] = ry;
  result = narray(new Num_Array(2, rshape, C, Malloc, rdata));

  return result;
}

