/* 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;  File: convolve.c
;;;  Author: Eero Simoncelli
;;;  Description: General convolution code for 2D images
;;;  Creation Date: Spring, 1986.
;;;  ----------------------------------------------------------------
;;;    Object-Based Vision and Image Understanding System (OBVIUS),
;;;      Copyright 1988, Vision Science Group,  Media Laboratory,  
;;;              Massachusetts Institute of Technology.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
*/

#include <stdio.h>
#include <math.h>


#define MAX_IM_SIZE 2048
#define abs(x)	  (x>=0 ? x : -(x))
#define ROOT2 1.4142135623730951
#define FILTER 0
#define EXPAND 1
#define IS    ==
#define ISNT  !=
#define AND &&
#define OR ||

typedef void (*fptr)();

typedef struct 
  {
  char *name;
  void (*func)();
  } EDGE_HANDLER;

/* 
This file contains functions which determine how edges are to be
handled when performing convolutions of images with linear filters.
Any edge handling function which is local and linear may be defined,
except (unfortunately) constants cannot be added.  So to treat the
edges as if the image is surrounded by a gray field, you must paste it
into a gray image, convolve, and crop it out...
The main convolution function is called internal_filter and is defined
in the file convolve.c.  The idea is that the convolution function
calls the edge handling function which computes a new filter based on
the old filter and the distance to the edge of the image.  For
example, reflection is done by reflecting the filter through the
appropriate axis and summing.  Currently defined functions are listed
below.  
*/

#define sgn(a)  ( ((a)>0)?1:(((a)<0)?-1:0) )
#define clip(a,mn,mx)  ( ((a)<(mn))?(mn):(((a)>=(mx))?(mx-1):(a)) )

static void reflect1(), reflect2(), repeat(), zero(), extend(), nocompute();
static void ereflect(), predict();

static EDGE_HANDLER edge_foos[] =
  {
    { "reflect1", reflect1 },   /* reflect about edge pixels - new name */
    { "reflect2", reflect2 },   /* standard reflection */
    { "repeat",   repeat   },   /* repeat edge pixel */
    { "zero",     zero     },   /* zero outside of image */
    { "extend",   extend   },   /* extend (reflect & invert) */
    { "dont-compute", nocompute }, /* ignore edge (zero output for filter near edge) */
    { "predict",  predict  },   /* predict based on portion covered by filt */
    { "ereflect", ereflect },	/* orthogonal QMF reflection */
    { "treflect", reflect1 }	/* reflect about edge pixels - old name */
  };

/*
Function looks up an edge handler id string in the structure above, and
returns the associated function 
*/

static fptr edge_function(edges)
  char *edges;
  {
  int i;
  for (i = 0; i<sizeof(edge_foos)/sizeof(EDGE_HANDLER); i++)
    if (!strcmp(edges,edge_foos[i].name))
      return(edge_foos[i].func);
  printf("No such edge handler routine!\n");
  return(0);
  }

/* 
---------------- EDGE HANDLER ARGUMENTS ------------------------
filt - floating point array of filter taps.
x_dim, y_dim - x and y dimensions of filt.
x_pos - position of filter relative to the horizontal image edges. Negative
       values indicate left edge, positive indicate right edge.  Zero 
       indicates that the filter is not touching either edge.  An absolute
       value of 1 indicates that the edge tap of the filter is over the 
       edge pixel of the image.
y_pos - analogous to x_pos.
result - floating point array where the resulting filter will go.  The edge
       of this filter will be aligned with the image for application...
f_or_e - equal to one of the two constants EXPAND or FILTER.
-------------------------------------------------------------------- 
*/ 


/* --------------------------------------------------------------------
zero() - Zero outside of image.  Discontinuous, but adds zero energy. */

static void
zero(filt,x_dim,y_dim,x_pos,y_pos,result,f_or_e)
  register float *filt, *result;
  register int x_dim;
  int y_dim, x_pos, y_pos, f_or_e;
  {
  register int y_filt,x_filt, y_res,x_res;
  int filt_sz = x_dim*y_dim;
  int x_start = ((x_pos>0)?(x_pos-1):((x_pos<0)?(x_pos+1):0));
  int y_start = x_dim * ((y_pos>0)?(y_pos-1):((y_pos<0)?(y_pos+1):0));
  int i;

  for (i=0; i<filt_sz; i++) result[i] = 0.0;

  for (y_filt=0, y_res=y_start;
       y_filt<filt_sz;
       y_filt+=x_dim, y_res+=x_dim)
      if ((y_res >= 0) AND (y_res < filt_sz))
	for (x_filt=y_filt, x_res=x_start;
	     x_filt<y_filt+x_dim;
	     x_filt++, x_res++)
	  if ((x_res >= 0) AND (x_res < x_dim))
	    result[y_res+x_res] = filt[x_filt];
  }

/* --------------------------------------------------------------------
repeat() - repeat edge pixel.  Continuous, but content is usually
different from image.  
*/
static void
repeat(filt,x_dim,y_dim,x_pos,y_pos,result,f_or_e)
  register float *filt, *result;
  register int x_dim;
  int y_dim, x_pos, y_pos, f_or_e;
  {
  register int y_filt,x_filt, y_res,x_res;
  int filt_sz = x_dim*y_dim;
  int x_start = ((x_pos>0)?(x_pos-1):((x_pos<0)?(x_pos+1):0));
  int y_start = x_dim * ((y_pos>0)?(y_pos-1):((y_pos<0)?(y_pos+1):0));
  int i;

  for (i=0; i<filt_sz; i++) result[i] = 0.0;

  for (y_filt=0, y_res=y_start;
       y_filt<filt_sz;
       y_filt+=x_dim, y_res+=x_dim)
    for (x_filt=y_filt, x_res=x_start;
	 x_filt<y_filt+x_dim;
	 x_filt++, x_res++)
      result[((y_res>=0)?((y_res<filt_sz)?y_res:(filt_sz-x_dim)):0) /* clip y_res */
	     + ((x_res>=0)?((x_res<x_dim)?x_res:(x_dim-1)):0)]      /* clip x_res */
	    += filt[x_filt];
  }

/* --------------------------------------------------------------------
reflect2() - "Normal" image reflection.  The edge pixel is repeated,
then the next pixel, etc.  Continuous, attempting to maintain
"similar" content, but discontinuous first derivative.
*/

static void
reflect2(filt,x_dim,y_dim,x_pos,y_pos,result,f_or_e)
  register float *filt, *result;
  register int x_dim;
  int y_dim, x_pos, y_pos, f_or_e;
  {
  register int y_filt,x_filt, y_edge,x_edge;
  register int x_base = (x_pos>0)?(x_dim-1):0;
  register int y_base = (y_pos>0)?(x_dim*(y_dim-1)):0; 
  int filt_sz = x_dim*y_dim;
  int x_edge_dist = (x_pos>0)?(x_pos-x_dim-1):(x_pos+1);
  int y_edge_dist = x_dim * ((y_pos>0)?(y_pos-y_dim-1):(y_pos+1));
  int i;

  for (i=0; i<filt_sz; i++) result[i] = 0.0;

  /* printf("x_edge_start: %d, x_base: %d\n", x_edge_dist, x_base); */
  for (y_filt=0, y_edge=y_edge_dist;
       y_filt<filt_sz;
       y_filt+=x_dim, y_edge+=x_dim)
      {
      if (y_edge IS 0) y_edge+=x_dim;
      for (x_filt=y_filt, x_edge=x_edge_dist;
	   x_filt<y_filt+x_dim;
	   x_filt++, x_edge++)
	  {
	  if (x_edge IS 0) x_edge++;
	  result[abs(y_base-abs(y_edge)+x_dim) + abs(x_base-abs(x_edge)+1)]
	    += filt[x_filt];
	  }
      }
  }

/* --------------------------------------------------------------------
reflect1() - Reflection through the edge pixels.  This is the right thing
to do if you are subsampling by 2, since it maintains parity (even 
pixels positions remain even, odd ones remain odd). (note: procedure differs 
depending on f_or_e parameter).  */	 

static void
reflect1(filt,x_dim,y_dim,x_pos,y_pos,result,f_or_e)
  register float *filt, *result;
  register int x_dim;
  int y_dim, x_pos, y_pos, f_or_e;
  {
  int filt_sz = x_dim*y_dim;
  register int x_start = 0, y_start = 0, x_stop = x_dim, y_stop = filt_sz;
  register int y_filt,x_filt, y_edge,x_edge;
  register int x_base = (x_pos>0)?(x_dim-1):0;
  register int y_base = (y_pos>0)?(x_dim*(y_dim-1)):0; 
  int x_edge_dist = (x_pos>0)?(x_pos-x_dim):((x_pos<-1)?(x_pos+1):0);
  int y_edge_dist = x_dim * ((y_pos>0)?(y_pos-y_dim):((y_pos<-1)?(y_pos+1):0));
  int i;
  int mx_pos = (x_dim/2)+1;
  int my_pos = (y_dim/2)+1;

  for (i=0; i<filt_sz; i++) result[i] = 0.0;

  /* if EXPAND and filter is centered on image edge, do not reflect */
  if (f_or_e IS EXPAND)
      {
      if (x_pos IS mx_pos) x_stop = (x_dim+1)/2;
      else if (x_pos IS -mx_pos) { x_start = x_dim/2; x_edge_dist = 0; }

      if (y_pos IS my_pos) y_stop = x_dim*((y_dim+1)/2);
      else if (y_pos IS -my_pos) { y_start = x_dim*(y_dim/2); y_edge_dist = 0;}
      }

  /* reflect at boundary of image */
  for (y_filt=y_start, y_edge=y_edge_dist;
       y_filt<y_stop;
       y_filt+=x_dim, y_edge+=x_dim)
    for (x_filt=y_filt+x_start, x_edge=x_edge_dist;
	 x_filt<y_filt+x_stop;
	 x_filt++, x_edge++)
      result[abs(y_base-abs(y_edge)) + abs(x_base-abs(x_edge))]
	+= filt[x_filt];

  /* if EXPAND and filter is not centered on image edge, mult edge by 2 */
  if (f_or_e IS EXPAND)
      {
      if ( (abs(x_pos) ISNT mx_pos) AND (x_pos ISNT 0) )
	for (y_filt=x_base; y_filt<filt_sz; y_filt+=x_dim)
	  result[y_filt] += result[y_filt];
      if ( (abs(y_pos) ISNT my_pos) AND (y_pos ISNT 0) )
	for (x_filt=y_base; x_filt<y_base+x_dim; x_filt++)
	  result[x_filt] += result[x_filt];
      }
  }

/* --------------------------------------------------------------------
extend() - Extend image by reflecting and inverting about edge pixel
value.  Maintains continuity in intensity AND first derivative (but
not higher derivs).
*/

static void
extend(filt,x_dim,y_dim,x_pos,y_pos,result,f_or_e)
  register float *filt, *result;
  register int x_dim;
  int y_dim, x_pos, y_pos, f_or_e;
  {
  int filt_sz = x_dim*y_dim;
  register int x_start = 0, y_start = 0, x_stop = x_dim, y_stop = filt_sz;
  register int y_filt,x_filt, y_edge,x_edge;
  register int x_base = (x_pos>0)?(x_dim-1):0;
  register int y_base = (y_pos>0)?(x_dim*(y_dim-1)):0; 
  int x_edge_dist = (x_pos>0)?(x_pos-x_dim):((x_pos<-1)?(x_pos+1):0);
  int y_edge_dist = x_dim * ((y_pos>0)?(y_pos-y_dim):((y_pos<-1)?(y_pos+1):0));
  int i;
  int mx_pos = (x_dim/2)+1;
  int my_pos = (y_dim/2)+1;

  for (i=0; i<filt_sz; i++) result[i] = 0.0;

  /* if EXPAND and filter is centered on image edge, do not reflect */
  if (f_or_e IS EXPAND)
      {
      if (x_pos IS mx_pos) x_stop = (x_dim+1)/2;
      else if (x_pos IS -mx_pos) { x_start = x_dim/2; x_edge_dist = 0; }

      if (y_pos IS my_pos) y_stop = x_dim*((y_dim+1)/2);
      else if (y_pos IS -my_pos) { y_start = x_dim*(y_dim/2); y_edge_dist = 0;}
      }

  /* reflect at boundary of image */
  for (y_filt=y_start, y_edge=y_edge_dist;
       y_filt<y_stop;
       y_filt+=x_dim, y_edge+=x_dim)
    for (x_filt=y_filt+x_start, x_edge=x_edge_dist;
	 x_filt<y_filt+x_stop;
	 x_filt++, x_edge++)
      if (((!y_base AND (sgn(y_edge) IS -1)) /* y overhanging */
	   OR
	   (y_base AND (sgn(y_edge) IS 1)))
	  ISNT			             /* XOR */
	  ((!x_base AND (sgn(x_edge) IS -1)) /* x overhanging */
	   OR
	   (x_base AND (sgn(x_edge) IS 1))))
	  {
	  result[abs(y_base-abs(y_edge)) + abs(x_base-abs(x_edge))]
	    -= filt[x_filt];
	  result[clip(y_base+y_edge,0,y_dim) + clip(x_base+x_edge,0,x_dim)]
	    += filt[x_filt] + filt[x_filt];
	  }
      else result[abs(y_base-abs(y_edge)) + abs(x_base-abs(x_edge))]
	  += filt[x_filt];

  }

/* --------------------------------------------------------------------
nocompute() - Return zero for values where filter hangs over the edge.
*/

static void
nocompute(filt,x_dim,y_dim,x_pos,y_pos,result,f_or_e)
  register float *filt, *result;
  register int x_dim;
  int y_dim, x_pos, y_pos, f_or_e;
  {
  register int i;
  register int size = x_dim*y_dim;

  if ( (x_pos>1) OR (x_pos<-1) OR (y_pos>1) OR (y_pos<-1) )
    for (i=0; i<size; i++)  result[i] = 0.0;
  else
    for (i=0; i<size; i++)  result[i] = filt[i];
  }

/* --------------------------------------------------------------------
Reflect, multiplying tap of filter which is at the edge of the image
by root 2.  This maintains orthogonality of QMF-style filters, but it
is not useful for most applications.  
*/

static void
ereflect(filt,x_dim,y_dim,x_pos,y_pos,result,f_or_e)
  register float *filt, *result;
  register int x_dim;
  int y_dim, x_pos, y_pos, f_or_e;
  {
  register int y_filt,x_filt, y_edge,x_edge;
  register int x_base = (x_pos>0)?(x_dim-1):0;
  register int y_base = x_dim * ( (y_pos>0)?(y_dim-1):0 ); 
  int filt_sz = x_dim*y_dim;
  int x_edge_dist = (x_pos>1)?(x_pos-x_dim):((x_pos<-1)?(x_pos+1):0);
  int y_edge_dist = x_dim * ( (y_pos>1)?(y_pos-y_dim):((y_pos<-1)?(y_pos+1):0) );
  int i;
  double norm,onorm;

  for (i=0; i<filt_sz; i++) result[i] = 0.0;

  /* reflect at boundary */       
  for (y_filt=0, y_edge=y_edge_dist;
       y_filt<filt_sz;
       y_filt+=x_dim, y_edge+=x_dim)
    for (x_filt=y_filt, x_edge=x_edge_dist;
	 x_filt<y_filt+x_dim;
	 x_filt++, x_edge++)
      result[abs(y_base-abs(y_edge)) + abs(x_base-abs(x_edge))]
	+= filt[x_filt];

  /* now multiply edge by root 2 */
  if (x_pos ISNT 0) 
    for (y_filt=x_base; y_filt<filt_sz; y_filt+=x_dim)
      result[y_filt] *= ROOT2;
  if (y_pos ISNT 0) 
    for (x_filt=y_base; x_filt<y_base+x_dim; x_filt++)
      result[x_filt] *= ROOT2;

  /* now normalize to norm of original filter */
  for (norm=0.0,i=0; i<filt_sz; i++)
    norm += (result[i]*result[i]);
  norm=sqrt(norm);

  for (onorm=0.0,i=0; i<filt_sz; i++)
    onorm += (filt[i]*filt[i]);
  onorm = sqrt(onorm);

  norm = norm/onorm;
  for (i=0; i<filt_sz; i++)
    result[i] /= norm;
  }

/* --------------------------------------------------------------------
predict() - version of prediction.  Like zero, but multiplies the result by
the reciprocal of the percentage of filter being used.  (i.e. if 50% of the
filter is hanging over the edge of the image, multiply the taps being used by
2.  *** Fix this to divide by the squared sum of taps being used and multiply by
the squared sum of ALL taps.
*/

static void
predict(filt,x_dim,y_dim,x_pos,y_pos,result,f_or_e)
  register float *filt, *result;
  register int x_dim;
  int y_dim, x_pos, y_pos, f_or_e;
  {
  register int y_filt,x_filt, y_res,x_res;
  register float taps_used = 0; /* int *** */
  register float fraction;
  int filt_sz = x_dim*y_dim;
  int x_start = ((x_pos>0)?(x_pos-1):((x_pos<0)?(x_pos+1):0));
  int y_start = x_dim * ((y_pos>0)?(y_pos-1):((y_pos<0)?(y_pos+1):0));
  int i;

  for (i=0; i<filt_sz; i++) result[i] = 0.0;

  for (y_filt=0, y_res=y_start;
       y_filt<filt_sz;
       y_filt+=x_dim, y_res+=x_dim)
      if ((y_res >= 0) AND (y_res < filt_sz))
	for (x_filt=y_filt, x_res=x_start;
	     x_filt<y_filt+x_dim;
	     x_filt++, x_res++)
	  if ((x_res >= 0) AND (x_res < x_dim))
	      {
	      result[y_res+x_res] = filt[x_filt];
	      taps_used += filt[x_filt]; /* 1.0 *** */
	      }

  if (f_or_e IS FILTER)
      {
      fraction = ( (float) 2.0 /* filt_sz *** */ ) / ( (float) taps_used );
      for (i=0; i<filt_sz; i++) result[i] *= fraction;
      }
  }


/* ------- printout stuff for testing ------------------------------
  fprintf(stderr,"Xpos: %d, Ypos: %d", x_pos, y_pos);
    for (y_filt=0; y_filt<y_dim; y_filt++)
      {
      fprintf(stderr,"\n");
      for (x_filt=0; x_filt<x_dim; x_filt++)
	fprintf(stderr,"%6.1f", result[y_filt*x_dim+x_filt]);
      }
  fprintf(stderr,"\n");
*/



/* Local Variables: */
/* buffer-read-only: t */
/* End: */

/* 
Performs correlation (i.e., convolution with filter(-x,-y))
followed by subsampling.  The operations are combined to avoid
unnecessary computation of the convolution samples that are to be
discarded in the subsampling operation.  The convolution is done in 9
sections so that mod operation is not performed unnecessarily.  The
grid_start parameters specify the starting location (first sample) of
the subsampling grid.  The grid_step parameters specify the step size
of the subsampling grid.  The stag parameter should be 0 or 1 and
indicates whether the subsampling grid is rectangular or quincunx (0
=> rectangular).  
*/

int
internal_wrap_filter(image,x_dim,y_dim,filt,x_fdim,y_fdim,
	      xgrid_start,xgrid_step,ygrid_start,ygrid_step,result,stag)
  register float *filt, *result;
  float *image;
  register int x_dim,y_dim,x_fdim,y_fdim;
  int xgrid_start,xgrid_step,ygrid_start,ygrid_step,stag;
  {
  register double sum;
  register int filt_size = x_fdim*y_fdim;
  float *imval[MAX_IM_SIZE];
  register int x_filt, x_im,y_im, y_filt_lin;
  register int compute = 1;
  int last_ctr_col = x_dim - x_fdim;
  int last_ctr_row = y_dim - y_fdim;
  int x_fmid = x_fdim/2;
  int y_fmid = y_fdim/2;
  int x_stop = x_dim - x_fmid;
  int y_stop = y_dim - y_fmid;
  int start_compute = stag - 1; 
  register int x_pos,y_pos, x_res;
  
  if (y_dim > MAX_IM_SIZE) 
      {
      printf("INTERNAL_WRAP: Image too large to filter!");
      return(1);
      }
  
  for (y_pos=y_im=0;y_pos<y_dim;y_pos++,y_im+=x_dim)
    imval[y_pos] = (image+y_im);
  
  for (y_pos=ygrid_start-y_fmid, x_res=0;           /* top */
       y_pos<0;
       y_pos+=ygrid_step)
      {
      if (stag)
	  {
	  start_compute = start_compute ? 0 : 1;    /* alternate on different lines */
	  compute = start_compute;
	  }
      for (x_pos=xgrid_start-x_fmid; 
	   x_pos<0;
	   x_pos+=xgrid_step,x_res++)
	  {
	  sum=0.0;
	  if (compute)
	    for (y_filt_lin=0,y_im=y_pos;
		 y_filt_lin<filt_size;
		 y_im++,y_filt_lin+=x_fdim)
	      for (x_im=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_im++)
		sum+=imval[(y_im+y_dim)%y_dim][(x_im+x_dim)%x_dim]*filt[x_filt];
	  result[x_res] = sum;
	  if (stag) compute = compute ? 0 : 1;      /* switch compute */
	  }
      for (; x_pos<last_ctr_col;
	   x_pos+=xgrid_step,x_res++) 
	  {
	  sum=0.0;
	  if (compute)
	    for (y_filt_lin=0,y_im=y_pos;
		 y_filt_lin<filt_size;
		 y_im++,y_filt_lin+=x_fdim)
	      for (x_im=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_im++)
		sum+=imval[(y_im+y_dim)%y_dim][x_im]*filt[x_filt];
	  result[x_res] = sum;
	  if (stag) compute = compute ? 0 : 1;
	  }
      for ( ; x_pos<x_stop;
	   x_pos+=xgrid_step,x_res++) 
	  {
	  sum=0.0;
	  if (compute)
	    for (y_filt_lin=0,y_im=y_pos;
		 y_filt_lin<filt_size;
		 y_im++,y_filt_lin+=x_fdim)
	      for (x_im=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_im++)
		sum+=imval[(y_im+y_dim)%y_dim][x_im%x_dim]*filt[x_filt];
	  result[x_res] = sum;
	  if (stag) compute = compute ? 0 : 1;
	  }
      }   
  
  for (;y_pos<last_ctr_row;	/* mid */
       y_pos+=ygrid_step)
      {
      if (stag)
	  {
	  start_compute = start_compute ? 0 : 1;    /* alternate on different lines */
	  compute = start_compute;
	  }
      for (x_pos=xgrid_start-x_fmid; 
	   x_pos<0;
	   x_pos+=xgrid_step,x_res++)
	  {
	  sum=0.0;
	  if (compute)
	    for (y_filt_lin=0,y_im=y_pos;
		 y_filt_lin<filt_size;
		 y_im++,y_filt_lin+=x_fdim)
	      for (x_im=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_im++)
		sum+=imval[y_im][(x_im+x_dim)%x_dim]*filt[x_filt];
	  result[x_res] = sum;
	  if (stag) compute = compute ? 0 : 1;
	  }
      for (; x_pos<last_ctr_col;
	   x_pos+=xgrid_step,x_res++) 
	  {
	  sum=0.0;
	  if (compute)
	    for (y_filt_lin=0,y_im=y_pos;
		 y_filt_lin<filt_size;
		 y_im++,y_filt_lin+=x_fdim)
	      for (x_im=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_im++)
		sum+=imval[y_im][x_im]*filt[x_filt];
	  result[x_res] = sum;
	  if (stag) compute = compute ? 0 : 1;
	  }
      for (; x_pos<x_stop;
	   x_pos+=xgrid_step,x_res++) 
	  {
	  sum=0.0;
	  if (compute)
	    for (y_filt_lin=0,y_im=y_pos;
		 y_filt_lin<filt_size;
		 y_im++,y_filt_lin+=x_fdim)
	      for (x_im=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_im++)
		sum+=imval[y_im][x_im%x_dim]*filt[x_filt];
	  result[x_res] = sum;
	  if (stag) compute = compute ? 0 : 1;
	  }
      }
  
  for (;y_pos<y_stop ;		/* bottom */
       y_pos+=ygrid_step) 
      {
      if (stag)
	  {
	  start_compute = start_compute ? 0 : 1;    /* alternate on different lines */
	  compute = start_compute;
	  }
      for (x_pos=xgrid_start-x_fmid; 
	   x_pos<0;
	   x_pos+=xgrid_step,x_res++)
	  {
	  sum=0.0;
	  if (compute)
	    for (y_filt_lin=0,y_im=y_pos;
		 y_filt_lin<filt_size;
		 y_im++,y_filt_lin+=x_fdim)
	      for (x_im=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_im++)
		sum+=imval[y_im%y_dim][(x_im+x_dim)%x_dim]*filt[x_filt];
	  result[x_res] = sum;
	  if (stag) compute = compute ? 0 : 1;
	  }
      for (; x_pos<last_ctr_col;
	   x_pos+=xgrid_step,x_res++) 
	  {
	  sum=0.0;
	  if (compute)
	    for (y_filt_lin=0,y_im=y_pos;
		 y_filt_lin<filt_size;
		 y_im++,y_filt_lin+=x_fdim)
	      for (x_im=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_im++)
		sum+=imval[y_im%y_dim][x_im]*filt[x_filt];
	  result[x_res] = sum;
	  if (stag) compute = compute ? 0 : 1;
	  }
      for (; x_pos<x_stop ;
	   x_pos+=xgrid_step,x_res++) 
	  {
	  sum=0.0;
	  if (compute)
	    for (y_filt_lin=0,y_im=y_pos;
		 y_filt_lin<filt_size;
		 y_im++,y_filt_lin+=x_fdim)
	      for (x_im=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_im++)
		sum+=imval[y_im%y_dim][x_im%x_dim]*filt[x_filt];
	  result[x_res] = sum;
	  if (stag) compute = compute ? 0 : 1;
	  }
      }     /* end bottom */
  return 0;
  }	/* end of internal_wrap_filter */



/* 
Performs upsampling (padding with zeros) followed by convolution.
The operations are combined to avoid unnecessary multiplication of
filter samples with zeros in the upsampled image.  The convolution is
done in 9 sections so that mod operation is not performed
unnecessarily.  Arguments are described in the comment above
internal_wrap_filter.
WARNING: this subroutine ADDS the expanded image into
the result, so the user must zero the result before invocation!  
*/

/* 3/20/95	butch	changed order of arguments to make it consistant */
int
internal_wrap_expand(image,x_dim, y_dim,filt,x_fdim,y_fdim,
	      xgrid_start,xgrid_step,ygrid_start,ygrid_step,result, stag)
  register float *filt, *result;
  float *image; 
  register int x_fdim,y_fdim,x_dim,y_dim;
  int xgrid_start,xgrid_step,ygrid_start,ygrid_step,stag;
  {
  register double val;
  register int filt_size = x_fdim*y_fdim;
  float *imval[MAX_IM_SIZE];
  register int x_filt, x_res,y_res, y_filt_lin;
  register int compute = 1;
  int last_ctr_col = x_dim - x_fdim;
  int last_ctr_row = y_dim - y_fdim;
  int x_fmid = x_fdim/2;
  int y_fmid = y_fdim/2;
  int x_stop = x_dim - x_fmid;
  int y_stop = y_dim - y_fmid;
  int start_compute = stag - 1; 
  register int x_pos,y_pos, x_im;
  
  if (y_dim > MAX_IM_SIZE) 
      {
      printf("INTERNAL_WRAP: Image too large to expand!");
      return(1);
      }
  
  for (y_pos=y_res=0;y_pos<y_dim;y_pos++,y_res+=x_dim)
    imval[y_pos] = (result+y_res);
  
  for (y_pos=ygrid_start-y_fmid, x_im=0;	/* top */
       y_pos<0;
       y_pos+=ygrid_step)
      {
      if (stag)
	  {
	  start_compute = start_compute ? 0 : 1;    /* alternate on different lines */
	  compute = start_compute;
	  }
      for (x_pos=xgrid_start-x_fmid; 
	   x_pos<0;
	   x_pos+=xgrid_step,x_im++)
	  {
	  val = image[x_im];
	  if (compute)
	    for (y_filt_lin=0,y_res=y_pos;
		 y_filt_lin<filt_size;
		 y_res++,y_filt_lin+=x_fdim)
	      for (x_res=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_res++)
		imval[(y_res+y_dim)%y_dim][(x_res+x_dim)%x_dim]
		  += val*filt[x_filt];
	  if (stag) compute = compute ? 0 : 1;      /* switch compute */
	  }
      for (; x_pos<last_ctr_col;
	   x_pos+=xgrid_step,x_im++) 
	  {
	  val = image[x_im];
	  if (compute)
	    for (y_filt_lin=0,y_res=y_pos;
		 y_filt_lin<filt_size;
		 y_res++,y_filt_lin+=x_fdim)
	      for (x_res=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_res++)
		imval[(y_res+y_dim)%y_dim][x_res]
		  += val*filt[x_filt];
	  if (stag) compute = compute ? 0 : 1;
	  }
      for ( ; x_pos<x_stop ;
	   x_pos+=xgrid_step,x_im++) 
	  {
	  val = image[x_im];		      
	  if (compute)
	    for (y_filt_lin=0,y_res=y_pos;
		 y_filt_lin<filt_size;
		 y_res++,y_filt_lin+=x_fdim)
	      for (x_res=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_res++)
		imval[(y_res+y_dim)%y_dim][x_res%x_dim]
		  += val*filt[x_filt];
	  if (stag) compute = compute ? 0 : 1;
	  }
      }   /* end top */
  
  for (;y_pos<last_ctr_row;	/* mid */
       y_pos+=ygrid_step)
      {
      if (stag)
	  {
	  start_compute = start_compute ? 0 : 1;    /* alternate on different lines */
	  compute = start_compute;
	  }
      for (x_pos=xgrid_start-x_fmid; 
	   x_pos<0;
	   x_pos+=xgrid_step,x_im++)
	  {
	  val = image[x_im];	 
	  if (compute)
	    for (y_filt_lin=0,y_res=y_pos;
		 y_filt_lin<filt_size;
		 y_res++,y_filt_lin+=x_fdim)
	      for (x_res=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_res++)
		imval[y_res][(x_res+x_dim)%x_dim]
		  += val*filt[x_filt];
	  if (stag) compute = compute ? 0 : 1;
	  }
      for (; x_pos<last_ctr_col;
	   x_pos+=xgrid_step,x_im++) 
	  {
	  val = image[x_im];
	  if (compute)
	    for (y_filt_lin=0,y_res=y_pos;
		 y_filt_lin<filt_size;
		 y_res++,y_filt_lin+=x_fdim)
	      for (x_res=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_res++)
		imval[y_res][x_res] += val*filt[x_filt];
	  if (stag) compute = compute ? 0 : 1;
	  }
      for (; x_pos<x_stop ;
	   x_pos+=xgrid_step,x_im++) 
	  {
	  val = image[x_im];		      
	  if (compute)
	    for (y_filt_lin=0,y_res=y_pos;
		 y_filt_lin<filt_size;
		 y_res++,y_filt_lin+=x_fdim)
	      for (x_res=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_res++)
		imval[y_res][x_res%x_dim]+= val*filt[x_filt];
	  if (stag) compute = compute ? 0 : 1;
	  }
      }             /* end mid */
  
  for (;y_pos<y_stop ;		/* bottom */
       y_pos+=ygrid_step) 
      {
      if (stag)
	  {
	  start_compute = start_compute ? 0 : 1;    /* alternate on different lines */
	  compute = start_compute;
	  }
      for (x_pos=xgrid_start-x_fmid; 
	   x_pos<0;
	   x_pos+=xgrid_step,x_im++)
	  {
	  val = image[x_im];
	  if (compute)
	    for (y_filt_lin=0,y_res=y_pos;
		 y_filt_lin<filt_size;
		 y_res++,y_filt_lin+=x_fdim)
	      for (x_res=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_res++)
		imval[y_res%y_dim][(x_res+x_dim)%x_dim]
		  += val*filt[x_filt];
	  if (stag) compute = compute ? 0 : 1;
	  }
      for (; x_pos<last_ctr_col;
	   x_pos+=xgrid_step,x_im++) 
	  {
	  val = image[x_im];		      
	  if (compute)
	    for (y_filt_lin=0,y_res=y_pos;
		 y_filt_lin<filt_size;
		 y_res++,y_filt_lin+=x_fdim)
	      for (x_res=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_res++)
		imval[y_res%y_dim][x_res] += val*filt[x_filt];
	  if (stag) compute = compute ? 0 : 1;
	  }
      for (; x_pos<x_stop;
	   x_pos+=xgrid_step,x_im++) 
	  {
	  val = image[x_im];
	  if (compute)
	    for (y_filt_lin=0,y_res=y_pos;
		 y_filt_lin<filt_size;
		 y_res++,y_filt_lin+=x_fdim)
	      for (x_res=x_pos,x_filt=y_filt_lin;
		   x_filt<y_filt_lin+x_fdim;
		   x_filt++,x_res++)
		imval[y_res%y_dim][x_res%x_dim]
		  += val*filt[x_filt];
	  if (stag) compute = compute ? 0 : 1;
	  }
      }          /* end bottom */
  return 0;
  }		/* end of internal_wrap_expand */



/* Local Variables: */
/* buffer-read-only: t */
/* End: */

/* --------------------------------------------------------------------
  Correlate filt with image, subsampling according to grid parameters,
  with values placed into result array.  Parameter temp is a temporary
  array the size of the filter.  edges is a string -- see convolve.h.
  The convolution is done in 9 sections, where the border sections use
  specially computed edge-handling filters (see edges.c).  
  The origin of the filter is assumed to be (x_dim/2, y_dim/2).
  Subsampling is done on a rectangular grid specified by the grid 
  parameters.  
  10/6/89 - approximately optimized the choice of register vars on SPARCS.
------------------------------------------------------------------------ */
int
internal_filter(image, x_dim, y_dim, filt, temp, x_fdim, y_fdim,
		xgrid_start,xgrid_step,ygrid_start,ygrid_step,result,edges)
  register float *image, *temp;
  register int x_fdim, x_dim;
  register int xgrid_step;
  register float *result;
  int ygrid_step;
  int xgrid_start, ygrid_start;
  float *filt; 
  int y_dim, y_fdim;
  char *edges;
  { 
  register double sum;
  register int x_filt, im_pos, y_filt_lin;
  register int y_im_lin, x_pos, filt_size = x_fdim*y_fdim;
  register int y_pos, res_pos;
  register int last_ctr_col = x_dim - x_fdim;
  int last_ctr_row = (y_dim - y_fdim) * x_dim;
  int first_row, first_col;
  int x_fmid = x_fdim/2;
  int y_fmid = y_fdim/2;
  int x_stop = x_fdim - x_fmid + 1;
  int y_stop = y_fdim - y_fmid + 1;
  int ygrid_step_full = ygrid_step*x_dim;
  int prev_res_pos, x_res_dim = (x_dim-xgrid_start+xgrid_step-1)/xgrid_step;
  int rt_edge_res_pos;
  fptr reflect = edge_function(edges);  /* look up edge-handling function */
  if (!reflect) 
      {
      fprintf(stderr,"Unknown edge handler: %s\n",edges);
      return(-1);
      }
  
  for (y_pos=ygrid_start-y_fmid-1,res_pos=0;    /* top */
       y_pos<0;
       y_pos+=ygrid_step)
      {
      for (x_pos=xgrid_start-x_fmid-1;        /* top-left corner */
	   x_pos<0;
	   x_pos+=xgrid_step,res_pos++)
	  {
	  (*reflect)(filt,x_fdim,y_fdim,x_pos,y_pos,temp,FILTER);
	  sum=0.0;
	  for (y_filt_lin=x_fdim,x_filt=y_im_lin=0;
	       y_filt_lin<=filt_size;
	       y_im_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (im_pos=y_im_lin;
		 x_filt<y_filt_lin;
		 x_filt++,im_pos++)
	      sum+=image[im_pos]*temp[x_filt];
	  result[res_pos] = sum;
	  }
      first_col = x_pos+1;
      (*reflect)(filt,x_fdim,y_fdim,0,y_pos,temp,FILTER);
      for (x_pos=first_col;	            /* top edge */
	   x_pos<last_ctr_col;
	   x_pos+=xgrid_step,res_pos++) 
	  {
	  sum=0.0;
	  for (y_filt_lin=x_fdim,x_filt=y_im_lin=0;
	       y_filt_lin<=filt_size;
	       y_im_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (im_pos=x_pos+y_im_lin;
		 x_filt<y_filt_lin;
		 x_filt++,im_pos++)
	      sum+=image[im_pos]*temp[x_filt];
	  result[res_pos] = sum;
	  }
      rt_edge_res_pos = res_pos + x_res_dim;   /* save this for later ... */
      for (x_pos+=(1-last_ctr_col);         /* top-right corner */
	   x_pos<x_stop;
	   x_pos+=xgrid_step,res_pos++) 
	  {
	  (*reflect)(filt,x_fdim,y_fdim,x_pos,y_pos,temp,FILTER);
	  sum=0.0;
	  for (y_filt_lin=x_fdim,x_filt=y_im_lin=0;
	       y_filt_lin<=filt_size;
	       y_im_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (im_pos=y_im_lin+last_ctr_col;
		 x_filt<y_filt_lin;
		 x_filt++,im_pos++)
	      sum+=image[im_pos]*temp[x_filt];
	  result[res_pos] = sum;
	  }
      }                        /* end top */   

  first_row = x_dim*(y_pos+1);   /* need this to go down the sides */
  prev_res_pos = res_pos;
  for (x_pos=xgrid_start-x_fmid-1;           /* left edge */
       x_pos<0;
       x_pos+=xgrid_step)
      {
      res_pos = prev_res_pos;
      (*reflect)(filt,x_fdim,y_fdim,x_pos,0,temp,FILTER);
      for (y_pos=first_row;
	   y_pos<last_ctr_row;
	   y_pos+=ygrid_step_full, res_pos+=x_res_dim)
	  {
	  sum=0.0;
	  for (y_filt_lin=x_fdim,x_filt=0,y_im_lin=y_pos;
	       y_filt_lin<=filt_size;
	       y_im_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (im_pos=y_im_lin;
		 x_filt<y_filt_lin;
		 x_filt++,im_pos++)
	      sum+=image[im_pos]*temp[x_filt];
	  result[res_pos] = sum;
	  }
      prev_res_pos++;
      }
  (*reflect)(filt,x_fdim,y_fdim,0,0,temp,FILTER);
  for (y_pos=first_row;		/* center region of image */
       y_pos<last_ctr_row;
       y_pos+=ygrid_step_full)
      {
      res_pos = prev_res_pos;
      for (x_pos=first_col;
	   x_pos<last_ctr_col;
	   x_pos+=xgrid_step,res_pos++) 
	  {
	  sum=0.0;
	  for (y_filt_lin=x_fdim,x_filt=0,y_im_lin=y_pos;
	       y_filt_lin<=filt_size;
	       y_im_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (im_pos=x_pos+y_im_lin;
		 x_filt<y_filt_lin;
		 x_filt++,im_pos++)
	      sum+=image[im_pos]*temp[x_filt];
	  result[res_pos] = sum;
	  }
      prev_res_pos+=x_res_dim;
      }
  prev_res_pos = rt_edge_res_pos;
  for (x_pos+=(1-last_ctr_col);                  /* right edge */
       x_pos<x_stop;
       x_pos+=xgrid_step) 
      {
      res_pos = prev_res_pos;
      (*reflect)(filt,x_fdim,y_fdim,x_pos,0,temp,FILTER);
      for (y_pos=first_row;
	   y_pos<last_ctr_row;
	   y_pos+=ygrid_step_full, res_pos+=x_res_dim)
	  {
	  sum=0.0;
	  for (y_filt_lin=x_fdim,x_filt=0,y_im_lin=y_pos;
	       y_filt_lin<=filt_size;
	       y_im_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (im_pos=y_im_lin+last_ctr_col;
		 x_filt<y_filt_lin;
		 x_filt++,im_pos++)
	      sum+=image[im_pos]*temp[x_filt];
	  result[res_pos] = sum;
	  }
      prev_res_pos++;
      }				/* end mid */
  res_pos -= (x_res_dim - 1);            /* go to lower left corner */
  for (y_pos=((y_pos-last_ctr_row)/x_dim)+1;     	/* bottom */
       y_pos<y_stop;
       y_pos+=ygrid_step) 
      {
      for (x_pos=xgrid_start-x_fmid-1;         /* bottom-left corner */
	   x_pos<0;
	   x_pos+=xgrid_step,res_pos++)
	  {
	  (*reflect)(filt,x_fdim,y_fdim,x_pos,y_pos,temp,FILTER);
	  sum=0.0;
	  for (y_filt_lin=x_fdim,x_filt=0,y_im_lin=last_ctr_row;
	       y_filt_lin<=filt_size;
	       y_im_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (im_pos=y_im_lin;
		 x_filt<y_filt_lin;
		 x_filt++,im_pos++)
	      sum+=image[im_pos]*temp[x_filt];
	  result[res_pos] = sum;
	  }
      (*reflect)(filt,x_fdim,y_fdim,0,y_pos,temp,FILTER);
      for (x_pos=first_col;		        /* bottom edge */
	   x_pos<last_ctr_col;
	   x_pos+=xgrid_step,res_pos++) 
	  {
	  sum=0.0;
	  for (y_filt_lin=x_fdim,x_filt=0,y_im_lin=last_ctr_row;
	       y_filt_lin<=filt_size;
	       y_im_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (im_pos=x_pos+y_im_lin;
		 x_filt<y_filt_lin;
		 x_filt++,im_pos++)
	      sum+=image[im_pos]*temp[x_filt];
	  result[res_pos] = sum;
	  }
      for (x_pos+=1-last_ctr_col;	/* bottom-right corner */
	   x_pos<x_stop;
	   x_pos+=xgrid_step,res_pos++) 
	  {
	  (*reflect)(filt,x_fdim,y_fdim,x_pos,y_pos,temp,FILTER);
	  sum=0.0;
	  for (y_filt_lin=x_fdim,x_filt=0,y_im_lin=last_ctr_row;
	       y_filt_lin<=filt_size;
	       y_im_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (im_pos=y_im_lin+last_ctr_col;
		 x_filt<y_filt_lin;
		 x_filt++,im_pos++)
	      sum+=image[im_pos]*temp[x_filt];
	  result[res_pos] = sum;
	  }
      }				/* end bottom */
  return 0;
  }			/* end of internal_filter */


/* --------------------------------------------------------------------

  Upsample image according to grid parameters and then convolve with
  filt, adding values into result array.

  WARNING: this subroutine ADDS the expanded image into the result, so
  the user must zero the result before invocation! 

------------------------------------------------------------------------ */	 
int
internal_expand(image,x_dim,y_dim,filt,temp,x_fdim,y_fdim,
				xgrid_start,xgrid_step,ygrid_start,ygrid_step,result,edges)
  register float *result, *temp;
  register int x_fdim, x_dim;
  register int xgrid_step;
  register float *image; 
  int ygrid_step;
  int xgrid_start, ygrid_start;
  float *filt; 
  int y_fdim, y_dim;
  char *edges;
  {
  register double val;
  register int x_filt, res_pos, y_filt_lin;
  register int y_res_lin, x_pos, filt_size = x_fdim*y_fdim;
  register int y_pos, im_pos;
  register int last_ctr_col = x_dim - x_fdim;
  int last_ctr_row = (y_dim - y_fdim) * x_dim;
  int first_col, first_row;
  int x_fmid = x_fdim/2;
  int y_fmid = y_fdim/2;
  int x_stop = x_fdim - x_fmid + 1;
  int y_stop = y_fdim - y_fmid + 1;
  int ygrid_step_full = ygrid_step*x_dim;
  int prev_im_pos, x_im_dim = (x_dim-xgrid_start+xgrid_step-1)/xgrid_step;
  int rt_edge_im_pos;
  fptr reflect = edge_function(edges);	 
  if (!reflect) 
      {
      fprintf(stderr,"Unknown edge handler: %s\n",edges);
      return(-1);
      }

  for (y_pos=ygrid_start-y_fmid-1,im_pos=0; /* top */
       y_pos<0;
       y_pos+=ygrid_step)
      {
      for (x_pos=xgrid_start-x_fmid-1; /* upper-left corner */
	   x_pos<0;
	   x_pos+=xgrid_step,im_pos++)
	  {
	  (*reflect)(filt,x_fdim,y_fdim,x_pos,y_pos,temp,EXPAND);
	  val = image[im_pos];
	  for (y_filt_lin=y_res_lin=0;
	       y_filt_lin<filt_size;
	       y_res_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (res_pos=y_res_lin,x_filt=y_filt_lin;
		 x_filt<y_filt_lin+x_fdim;
		 x_filt++,res_pos++)
	      result[res_pos] += val*temp[x_filt];
	  }
      first_col = x_pos+1;
      (*reflect)(filt,x_fdim,y_fdim,0,y_pos,temp,EXPAND);
      for (x_pos=first_col;		/* top edge */
	   x_pos<last_ctr_col;
	   x_pos+=xgrid_step,im_pos++) 
	  {
	  val = image[im_pos];
	  for (y_filt_lin=y_res_lin=0;
	       y_filt_lin<filt_size;
	       y_res_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (res_pos=x_pos+y_res_lin,x_filt=y_filt_lin;
		 x_filt<y_filt_lin+x_fdim;
		 x_filt++,res_pos++)
	      result[res_pos]  += val*temp[x_filt];
	  }
      rt_edge_im_pos = im_pos + x_im_dim;
      for (x_pos+=(1-last_ctr_col);	/* upper-right corner */
	   x_pos<x_stop;
	   x_pos+=xgrid_step,im_pos++) 
	  {
	  (*reflect)(filt,x_fdim,y_fdim,x_pos,y_pos,temp,EXPAND);
	  val = image[im_pos];		      
	  for (y_filt_lin=y_res_lin=0;
	       y_filt_lin<filt_size;
	       y_res_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (res_pos=y_res_lin+last_ctr_col,x_filt=y_filt_lin;
		 x_filt<y_filt_lin+x_fdim;
		 x_filt++,res_pos++)
	      result[res_pos] += val*temp[x_filt];
	  }
      }				    /* end top */   
  
  first_row = x_dim*(y_pos+1);      /* need this to go down sides */
  prev_im_pos = im_pos;
  for (x_pos=xgrid_start-x_fmid-1;  /* left edge */
       x_pos<0;
       x_pos+=xgrid_step)
      {
      im_pos = prev_im_pos;
      (*reflect)(filt,x_fdim,y_fdim,x_pos,0,temp,EXPAND);
      for (y_pos=first_row;
	   y_pos<last_ctr_row;                
	   y_pos+=ygrid_step_full, im_pos+=x_im_dim)
	  {
	  val = image[im_pos];
	  for (y_filt_lin=0,y_res_lin=y_pos;
	       y_filt_lin<filt_size;
	       y_res_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (res_pos=y_res_lin,x_filt=y_filt_lin;
		 x_filt<y_filt_lin+x_fdim;
		 x_filt++,res_pos++)
	      result[res_pos] += val*temp[x_filt];
	  }
      prev_im_pos++;
      }
  (*reflect)(filt,x_fdim,y_fdim,0,0,temp,EXPAND);
  for (y_pos=first_row;		/* center region of image */
       y_pos<last_ctr_row;
       y_pos+=ygrid_step_full)
      {
      im_pos = prev_im_pos;
      for (x_pos=first_col;
	   x_pos<last_ctr_col;
	   x_pos+=xgrid_step,im_pos++) 
	  {
	  val = image[im_pos];
	  for (y_filt_lin=0,y_res_lin=y_pos;
	       y_filt_lin<filt_size;
	       y_res_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (res_pos=x_pos+y_res_lin,x_filt=y_filt_lin;
		 x_filt<y_filt_lin+x_fdim;
		 x_filt++,res_pos++)
	      result[res_pos]  += val*temp[x_filt];
	  }
      prev_im_pos+=x_im_dim;
      }
  prev_im_pos = rt_edge_im_pos;
  for (x_pos+=(1-last_ctr_col);	/* right edge */
       x_pos<x_stop;
       x_pos+=xgrid_step)
      {
      im_pos = prev_im_pos;
      (*reflect)(filt,x_fdim,y_fdim,x_pos,0,temp,EXPAND);
      for (y_pos=first_row;
	   y_pos<last_ctr_row;
	   y_pos+=ygrid_step_full, im_pos+=x_im_dim)
	  {
	  val = image[im_pos];		      
	  for (y_filt_lin=0,y_res_lin=y_pos;
	       y_filt_lin<filt_size;
	       y_res_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (res_pos=y_res_lin+last_ctr_col,x_filt=y_filt_lin;
		 x_filt<y_filt_lin+x_fdim;
		 x_filt++,res_pos++)
	      result[res_pos] += val*temp[x_filt];
	  }
      prev_im_pos++;
      }				/* end mid */
  
  im_pos -= (x_im_dim -1);
  for (y_pos=((y_pos-last_ctr_row)/x_dim)+1;
       y_pos<y_stop;		/* bottom */
       y_pos+=ygrid_step) 
      {
      for (x_pos=xgrid_start-x_fmid-1; 
	   x_pos<0;
	   x_pos+=xgrid_step,im_pos++)
	  {
	  (*reflect)(filt,x_fdim,y_fdim,x_pos,y_pos,temp,EXPAND);
	  val = image[im_pos];
	  for (y_filt_lin=0,y_res_lin=last_ctr_row;
	       y_filt_lin<filt_size;
	       y_res_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (res_pos=y_res_lin,x_filt=y_filt_lin;
		 x_filt<y_filt_lin+x_fdim;
		 x_filt++,res_pos++)
	      result[res_pos] += val*temp[x_filt];
	  }
      (*reflect)(filt,x_fdim,y_fdim,0,y_pos,temp,EXPAND);
      for (x_pos=first_col;
	   x_pos<last_ctr_col;
	   x_pos+=xgrid_step,im_pos++) 
	  {
	  val = image[im_pos];
	  for (y_filt_lin=0, y_res_lin=last_ctr_row;
	       y_filt_lin<filt_size;
	       y_res_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (res_pos=x_pos+y_res_lin,x_filt=y_filt_lin;
		 x_filt<y_filt_lin+x_fdim;
		 x_filt++,res_pos++)
	      result[res_pos]  += val*temp[x_filt];
	  }
      for (x_pos+=1-last_ctr_col;
	   x_pos<x_stop;
	   x_pos+=xgrid_step,im_pos++) 
	  {
	  (*reflect)(filt,x_fdim,y_fdim,x_pos,y_pos,temp,EXPAND);
	  val = image[im_pos];		      
	  for (y_filt_lin=0,y_res_lin=last_ctr_row;
	       y_filt_lin<filt_size;
	       y_res_lin+=x_dim,y_filt_lin+=x_fdim)
	    for (res_pos=y_res_lin+last_ctr_col,x_filt=y_filt_lin;
		 x_filt<y_filt_lin+x_fdim;
		 x_filt++,res_pos++)
	      result[res_pos] += val*temp[x_filt];
	  }
      }				/* end bottom */
  return 0;
  }				/* end of internal_expand */
  

/* Local Variables: */
/* buffer-read-only: t */
/* End: */

