/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 *
 * This code is a derivative work of other copyrighted software.  See the
 * file COPYRIGHTS for more information on terms and conditions.
 */

extern "C" {
#include <stk.h>
}
#include <narray.hh>

/****************************************************************************
 * DCT	--- Discrete cosine transformation
 */

#define ONE_ON_ROOT_TWO	0.707106781188
#define PI		3.1415926536

static float *
sgen_dct_matrix(int size, int type)
{
    float *matrix = (float *)malloc(size * size * sizeof(float));
    
    if(matrix != NULL) {
	int e=0;
	float k=sqrt(2.0 / size);
	for(int i=0; i<size; i++) {
	    for(int j=0; j<size; j++, e++) {
		matrix[e]  = k;
		switch(type) {
		case 2:
		    matrix[e] *= (i==0 ? ONE_ON_ROOT_TWO : 1.0);
		    matrix[e] *= cos(((2.0 * j + 1) * i * PI) / (2.0 * size));
		    break;
		case 3:
		    matrix[e] *= (j==0 ? ONE_ON_ROOT_TWO : 1.0);
		    matrix[e] *= cos(((2.0 * i + 1) * j * PI) / (2.0 * size));
		    break;
		case 4:
		    matrix[e] *= cos(((2.0 * i + 1) * (2.0 * j + 1) * PI) 
				     / (4.0 * size));
		    break;
		}
	    }
	}
    }
    return matrix;
}

	    
static double *
dgen_dct_matrix(int size, int type)
{
    double *matrix = (double *)malloc(size * size * sizeof(double));
    
    if(matrix != NULL) {
	int e=0;
	for(int i=0; i<size; i++) {
	    for(int j=0; j<size; j++, e++) {
		matrix[e]  = sqrt(2.0 / size);
		switch(type) {
		case 2:
		    matrix[e] *= (i==0 ? 1.0 : ONE_ON_ROOT_TWO);
		    matrix[e] *= cos(((2.0 * j + 1) * i * PI) / (2.0 * size));
		    break;
		case 3:
		    matrix[e] *= (j==0 ? 1.0 : ONE_ON_ROOT_TWO);
		    matrix[e] *= cos(((2 * i + 1) * j * PI) / (2 * size));
		    break;
		case 4:
		    matrix[e] *= cos(((2.0 * i + 1) * (2.0 * j + 1) * PI) 
				     / (4.0 * size));
		    break;
		}
	    }
	}
    }
    return matrix;
}

static _Num_Array *
dct_1D_transform(const _Num_Array *vector, _Num_Array *result, int type)
/*
 * Note: vector must be 1D; dct types are described above
 *	'result' is required to be an abstract numeric array with the same
 *	dimension (1) and shape as 'vector'
 */
{
    static float *raw_dct_matrix = NULL;
    static int current_type = 0;
    static index_type current_size;

    index_type ind[MAX_DIMENSION];
    index_type size = vector->shape_ref(0);
    T *res;

    memset(ind, 0, MAX_DIMENSION*sizeof(index_type));

    if(raw_dct_matrix == NULL 
       || type != current_type
       || size != current_size) {
	if(raw_dct_matrix != NULL)
	    free(raw_dct_matrix);
	raw_dct_matrix = sgen_dct_matrix(size, type);
	current_type = type;
	current_size = size;
    }

    for(index_type i=0; i<size; i++) {
	ind[0] = i;
	res = result->elem_iref(ind);
	*res = 0;
	for(index_type j=0; j<size; j++) {
	    ind[0] = j;
	    *res += vector->elem_ref(ind) * raw_dct_matrix[i*size+j];
	}
    }
    return result;
}


/****************************************************************************
 * Utility functions
 */


// static void
// rotate_shape_left(_Num_Array& array)
// {
//     dim_type new_order[MAX_DIMENSION];
//     new_order[array.dimension_ref()-1] = 0;
//     for(dim_type i=1; i<array.dimension_ref(); i++)
// 	new_order[i-1] = i;
//     array.permute(new_order);
// }

static Num_Array *
apply_separable_transform (const Num_Array *indata, 
			   _Num_Array *(*stf)(const _Num_Array *indata,
					      _Num_Array *outdata,
					      int type),
			   int type)
{
    dim_type d, ndim = indata->dimension_ref();
    index_type block_shape[MAX_DIMENSION];
    for(d=0; d<ndim; d++)
	block_shape[d] = 1;

    Num_Array	 *result;
    Num_Array	 *_indata = (Num_Array *)indata;
    Num_Array	 *tmp_data = NULL;
    Num_Subarray *inblock, *outblock;
    int nblocks;
    
    for(d=0; d<indata->dimension_ref(); d++) {
	tmp_data = new Num_Array(_indata->dimension_ref(), 
				 _indata->shape_ref());
	block_shape[0] = _indata->shape_ref(0);
	inblock = new Num_Subarray(*_indata, block_shape);
	outblock = new Num_Subarray(*tmp_data, block_shape);
	nblocks = indata->size_ref() / block_shape[0];
	for(index_type i=0; i<nblocks; i++) {
	    stf(inblock, outblock, type);
	    inblock->inc_position();
	    outblock->inc_position();
	}
	delete inblock;
	delete outblock;
	if(_indata != indata)
	    delete _indata;
	_indata = tmp_data;
	result = tmp_data;
	_indata->rotate();
    }
    return result;
}

/****************************************************************************
 * Work functions
 */

Num_Array *
ND_dct_transform(Num_Array *indata, int type)
{
    return apply_separable_transform(indata, dct_1D_transform, type);
}

/****************************************************************************
 * STk Wrapper functions
 */

extern "C" PRIMITIVE
narray_dct(SCM indata, SCM type)
{
    if(! NARRAYP(indata))
	STk_err("narray-dct: not an narray", indata);
    if(!INTEGERP(type) || INTEGER(type) < 2 || INTEGER(type) > 4)
	STk_err("narray-dct: not a valid DCT type", type);

    return narray(ND_dct_transform(narray_instance(indata), INTEGER(type)));
}
