/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 *
 * This code is a derivative work of other copyrighted software.  See the
 * file COPYRIGHTS for more information on terms and conditions.
 */

/*
 * morphology.hh
 *
 * The following is a set of templated morphology routines.  All of
 * these routines use reflection about the edge points to calculate
 * boundary results.  For this reason the result array is assumed to
 * be the same size as the data array.  The result is centered if the
 * kerenel dimension is odd, left of center if the kernel length is
 * even.  Both binary and grescale versions are given but the binary
 * versions are currently unable to operate on packed bit-arrays.  
 */

#define MAX(a,b)	((a)>(b) ? (a) : (b))
#define MIN(a,b)	((a)<(b) ? (a) : (b))

template <class DT>
inline DT
reflected_access(const DT *data, int dx, int dy, int x, int y)
{
  int rx, ry;
  rx = ((x>0) ? ((x>=dx) ? (dx<<1)-x-2 : x) : -x);
  ry = ((y>0) ? ((y>=dy) ? (dy<<1)-y-2 : y) : -y);

  return data[rx + ry*dx];
}

template <class DT, class KT>
void
binary_dilate_reflected(DT *result,
		       const DT *data, int dx, int dy,
		       const KT *kern, int kx, int ky)
{
  const DT *d, *b, *br, *k;
  DT *r, tmp;
  int xd, yd, xr, yr, xk, yk;		/* data,result,kernel indices	*/
  int lbw = (int)floor((float)kx/2.0);	/* left border width		*/
  int tbw = (int)floor((float)ky/2.0);	/* top border width		*/
  int rbs = lbw + dx - kx + 1;		/* right border start		*/
  int bbs = tbw + dy - ky + 1;		/* bottom border start		*/

  memset(result,0,dx*dy*sizeof(DT));

  /* top	*/
  for(yr=0,yd=-tbw; yd<0; yr++,yd++)
    for(xr=0,xd=-lbw; xr<dx; xr++,xd++)
      for(yk=0,r=&result[xr+yr*dx],k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++)
	  if((*k) != 0) {
	    tmp = reflected_access(data,dx,dy,xd+xk,yd+yk);
	    *r  = MAX(*r, tmp);
	  }
  /* middle	*/
  for(; yr<bbs; yr++,yd++) {
    /* left border	*/
    for(xr=0,xd=-lbw; xd<0; xr++,xd++)
      for(yk=0,r=&result[xr+yr*dx],k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++)
	  if((*k) != 0) {
	    tmp = reflected_access(data,dx,dy,xd+xk,yd+yk);
	    *r  = MAX(*r, tmp);
	  }
    /* center	(uses more efficient array access in loop) */
    for(d=&data[yd*dx], r=&result[xr+yr*dx]; xr<rbs; xr++,xd++,d++,r++)
      for(k=kern, yk=0, br=d;  yk<ky  ;yk++, br+=dx)
	for(xk=0, b=br; xk<kx; xk++, b++, k++)
	  (*r)=((*k)!=0 ? MAX(*r,*b) : (*r));
    /* right border	*/
    for(; xr<dx; xr++,xd++)
      for(yk=0,r=&result[xr+yr*dx],k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++)
	  if((*k) != 0) {
	    tmp = reflected_access(data,dx,dy,xd+xk,yd+yk);
	    *r  = MAX(*r, tmp);
	  }
  }
  /* bottom	*/
  for(; yr<dy; yr++,yd++)
    for(xr=0,xd=-lbw; xr<dx; xr++,xd++)
      for(yk=0,r=&result[xr+yr*dx],k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++)
	  if((*k) != 0) {
	    tmp = reflected_access(data,dx,dy,xd+xk,yd+yk);
	    *r  = MAX(*r, tmp);
	  }
}
  
template <class DT, class KT>
void
grey_dilate_reflected(DT *result,
		      const DT *data, int dx, int dy,
		      const KT *kern, int kx, int ky)
{
  const DT *d, *b, *br, *k;
  DT *r, tmp;
  int xd, yd, xr, yr, xk, yk;		/* data,result,kernel indices	*/
  int lbw = (int)floor((float)kx/2.0);	/* left border width		*/
  int tbw = (int)floor((float)ky/2.0);	/* top border width		*/
  int rbs = lbw + dx - kx + 1;		/* right border start		*/
  int bbs = tbw + dy - ky + 1;		/* bottom border start		*/

  memset(result,0,dx*dy*sizeof(DT));

  /* top	*/
  for(yr=0,yd=-tbw; yd<0; yr++,yd++)
    for(xr=0,xd=-lbw; xr<dx; xr++,xd++)
      for(yk=0,r=&result[xr+yr*dx],k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++) {
	  tmp = (*k) + reflected_access(data,dx,dy,xd+xk,yd+yk);
	  (*r) = MAX((*r),tmp);
	}
  /* middle	*/
  for(; yr<bbs; yr++,yd++) {
    /* left border	*/
    for(xr=0,xd=-lbw; xd<0; xr++,xd++)
      for(yk=0,r=&result[xr+yr*dx],k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++) {
	  tmp = (*k) + reflected_access(data,dx,dy,xd+xk,yd+yk);
	  (*r) = MAX((*r),tmp);
	}
    /* center	(uses more efficient array access in loop) */
    for(d=&data[yd*dx], r=&result[xr+yr*dx]; xr<rbs; xr++,xd++,d++,r++)
      for(k=kern, yk=0, br=d;  yk<ky  ;yk++, br+=dx)
	for(xk=0, b=br; xk<kx; xk++, b++, k++) {
	  tmp = (*k) + (*b);
	  (*r)= MAX((*r),tmp);
	}
    /* right border	*/
    for(; xr<dx; xr++,xd++)
      for(yk=0,r=&result[xr+yr*dx],k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++) {
	  tmp = (*k) + reflected_access(data,dx,dy,xd+xk,yd+yk);
	  (*r) = MAX((*r),tmp);
	}
  }
  /* bottom	*/
  for(; yr<dy; yr++,yd++)
    for(xr=0,xd=-lbw; xr<dx; xr++,xd++)
      for(yk=0,r=&result[xr+yr*dx],k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++) {
	  tmp = (*k) + reflected_access(data,dx,dy,xd+xk,yd+yk);
	  (*r) = MAX((*r),tmp);
	}
}
	     
template <class DT, class KT>
void
binary_erode_reflected(DT *result,
		       const DT *data, int dx, int dy,
		       const KT *kern, int kx, int ky)
{
  const DT *d, *b, *br, *k;
  DT *r, max, tmp;
  int xd, yd, xr, yr, xk, yk;		/* data,result,kernel indices	*/
  int lbw = (int)floor((float)kx/2.0);	/* left border width		*/
  int tbw = (int)floor((float)ky/2.0);	/* top border width		*/
  int rbs = lbw + dx - kx + 1;		/* right border start		*/
  int bbs = tbw + dy - ky + 1;		/* bottom border start		*/

  for(max=(*data),xd=0,d=data; xd<dx*dy; xd++,d++)
    if((*d) > max)
      max = *d;
    
  /* top	*/
  for(yr=0,yd=-tbw; yd<0; yr++,yd++)
    for(xr=0,xd=-lbw; xr<dx; xr++,xd++) {
      r = &result[xr+yr*dx];
      for(yk=0, *r=max, k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++)
	  if((*k) != 0) {
	    tmp = reflected_access(data,dx,dy,xd+xk,yd+yk);
	    *r  = MIN(*r, tmp);
	  }
    }
  /* middle	*/
  for(; yr<bbs; yr++,yd++) {
    /* left border	*/
    for(xr=0,xd=-lbw; xd<0; xr++,xd++) {
      r = &result[xr+yr*dx];
      for(yk=0, *r=max, k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++)
	  if((*k) != 0) {
	    tmp = reflected_access(data,dx,dy,xd+xk,yd+yk);
	    *r  = MIN(*r, tmp);
	  }
    }
    /* center	(uses more efficient array access in loop) */
    for(d=&data[yd*dx], r=&result[xr+yr*dx]; xr<rbs; xr++,xd++,d++,r++)
      for(*r=max, k=kern, yk=0, br=d;  yk<ky  ;yk++, br+=dx)
	for(xk=0, b=br; xk<kx; xk++, b++, k++)
	  (*r)=((*k)!=0 ? MIN(*r,*b) : (*r));
    /* right border	*/
    for(; xr<dx; xr++,xd++) {
      r = &result[xr+yr*dx];
      for(yk=0, *r=max, k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++)
	  if((*k) != 0) {
	    tmp = reflected_access(data,dx,dy,xd+xk,yd+yk);
	    *r  = MIN(*r, tmp);
	  }
    }
  }
  /* bottom	*/
  for(; yr<dy; yr++,yd++)
    for(xr=0,xd=-lbw; xr<dx; xr++,xd++) {
      r = &result[xr+yr*dx];
      for(yk=0, *r=max, k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++)
	  if((*k) != 0) {
	    tmp = reflected_access(data,dx,dy,xd+xk,yd+yk);
	    *r  = MIN(*r, tmp);
	  }
    }
}
  
template <class DT, class KT>
void
grey_erode_reflected(DT *result,
		      const DT *data, int dx, int dy,
		      const KT *kern, int kx, int ky)
{
  const DT *d, *b, *br, *k;
  DT *r, tmp, max;
  int xd, yd, xr, yr, xk, yk;		/* data,result,kernel indices	*/
  int lbw = (int)floor((float)kx/2.0);	/* left border width		*/
  int tbw = (int)floor((float)ky/2.0);	/* top border width		*/
  int rbs = lbw + dx - kx + 1;		/* right border start		*/
  int bbs = tbw + dy - ky + 1;		/* bottom border start		*/

  for(max=(*data),xd=0,d=data; xd<dx*dy; xd++,d++)
    if((*d) > max)
      max = *d;

  /* top	*/
  for(yr=0,yd=-tbw; yd<0; yr++,yd++)
    for(xr=0,xd=-lbw; xr<dx; xr++,xd++) {
      r = &result[xr+yr*dx];
      for(yk=0, *r=max, k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++) {
	  tmp = -(*k) + reflected_access(data,dx,dy,xd+xk,yd+yk);
	  (*r) = MIN((*r),tmp);
	}
    }
  /* middle	*/
  for(; yr<bbs; yr++,yd++) {
    /* left border	*/
    for(xr=0,xd=-lbw; xd<0; xr++,xd++) {
      r = &result[xr+yr*dx];
      for(yk=0, *r=max, k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++) {
	  tmp = -(*k) + reflected_access(data,dx,dy,xd+xk,yd+yk);
	  (*r) = MIN((*r),tmp);
	}
    }
    /* center	(uses more efficient array access in loop) */
    for(d=&data[yd*dx], r=&result[xr+yr*dx]; xr<rbs; xr++,xd++,d++,r++)
      for(*r=max, k=kern, yk=0, br=d;  yk<ky  ;yk++, br+=dx)
	for(xk=0, b=br; xk<kx; xk++, b++, k++) {
	  tmp = -(*k) + (*b);
	  (*r)= MIN((*r),tmp);
	}
    /* right border	*/
    for(; xr<dx; xr++,xd++) {
      r = &result[xr+yr*dx];
      for(yk=0, *r=max, k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++) {
	  tmp = -(*k) + reflected_access(data,dx,dy,xd+xk,yd+yk);
	  (*r) = MIN((*r),tmp);
	}
    }
  }
  /* bottom	*/
  for(; yr<dy; yr++,yd++)
    for(xr=0,xd=-lbw; xr<dx; xr++,xd++) {
      r = &result[xr+yr*dx];
      for(yk=0, *r=max, k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++) {
	  tmp = -(*k) + reflected_access(data,dx,dy,xd+xk,yd+yk);
	  (*r) = MIN((*r),tmp);
	}
    }
}
	     

template <class RT, class AT, class BT>
void
elementwise_infimum(RT *result, AT *A, BT *B, ulong n)
{
  for(;n>0; n--, result++, A++, B++)
    *result = MIN(*A, *B);
}

template <class RT, class AT, class BT>
void
elementwise_supremum(RT *result, AT *A, BT *B, ulong n)
{
  for(;n>0; n--, result++, A++, B++)
    *result = MAX(*A, *B);
}


template <class DT, class KT>
void
binary_erode_reflected2(DT *result,
			const DT *data, int dx, int dy,
			const KT *kern, int kx, int ky)
{
  const DT *d, *b, *br, *k;
  DT *r, tmp, max;
  int xd, yd, xr, yr, xk, yk;		/* data,result,kernel indices	*/
  int lbw = (int)floor((float)kx/2.0);	/* left border width		*/
  int tbw = (int)floor((float)ky/2.0);	/* top border width		*/
  int rbs = lbw + dx - kx + 1;		/* right border start		*/
  int bbs = tbw + dy - ky + 1;		/* bottom border start		*/

  for(max=(*data),xd=0,d=data; xd<dx*dy; xd++,d++)
    if((*d) > max)
      max = *d;

  for(yr=0, yd=-tbw; yr<dy; yr++, yd++)
    for(xr=0,xd=-lbw; xr<dx; xr++,xd++) {
      r = &result[xr+yr*dx];
      for(yk=0, *r=max, k=kern; yk<ky; yk++)
	for(xk=0; xk<kx; xk++,k++)
	  if((*k)!=0) {
	    tmp = reflected_access(data,dx,dy,xd+xk,yd+yk);
	    *r  = MIN(*r, tmp);
	  }
    }
}
