/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

#ifndef __IND_ARRAY__
#define __IND_ARRAY__
/*
 * IND_ARRAY.hh	--- indirect arrays
 *
 * This defines the base classes for all arrays which are indirect.  That is
 * they do not have their own data but reference the data in another array
 * in some useful way.  The most significant examples include:
 *	Subarray:	reference an area within a parent array
 *	Ext_Array:	allow out-of_bounds array reference through use of
 *			a structured extension (such as periodisation).
 */

#include <ND_Array.hh>

class _Indirect_Array
    : public virtual _ND_Array
{
protected:
    _ND_Array* parent;

public:
    _Indirect_Array() : parent(0) {;}
    _Indirect_Array(const _ND_Array& the_parent) { 
	parent = (_ND_Array *)&the_parent; 
    }
    virtual ~_Indirect_Array() {;}
    _ND_Array* parent_ref() { return parent; }
    void parent_set(const _ND_Array& new_parent) { 
	parent = (_ND_Array *)&new_parent; 
    }
    dim_type dimension_ref() const { return parent->dimension_ref(); }
    const index_type* shape_ref() const { return parent->shape_ref(); }
    index_type shape_ref(dim_type d) const { return parent->shape_ref(d); }
    index_type *inc_index(index_type *ind, int step = 1) const
    {
	return parent->inc_index(ind, step);
    }
};

class _Indirect_Numarray
: public virtual _Num_Array
{
  protected:
    _Num_Array *parent;
  public:
    _Indirect_Numarray() : parent(0) {;}
    _Indirect_Numarray(const _Num_Array& the_parent) {
	parent = (_Num_Array *)&the_parent; 
    }
    virtual ~_Indirect_Numarray() {;}

    _Num_Array* parent_ref() { return parent; }
    void parent_set(const _Num_Array& new_parent) { 
	parent = (_Num_Array *)&new_parent; 
    }
    dim_type dimension_ref() const { return parent->dimension_ref(); }
    const index_type* shape_ref() const { return parent->shape_ref(); }
    index_type shape_ref(dim_type d) const { return parent->shape_ref(d); }
    index_type *inc_index(index_type *ind, int step = 1) const
    {
	return parent->inc_index(ind, step);
    }
};

class Indirect_Numarray
    : public virtual _Num_Array
{
  protected:
    Num_Array *parent;
  public:
    Indirect_Numarray() : parent(0) {;}
    Indirect_Numarray(const Num_Array& the_parent) {
	parent = (Num_Array *)&the_parent; 
    }
    virtual ~Indirect_Numarray() {;}

    Num_Array* parent_ref() { return parent; }
    void parent_set(const Num_Array& new_parent) { 
	parent = (Num_Array *)&new_parent; 
    }
    dim_type dimension_ref() const { return parent->dimension_ref(); }
    const index_type* shape_ref() const { return parent->shape_ref(); }
    index_type shape_ref(dim_type d) const { return parent->shape_ref(d); }
    index_type *inc_index(index_type *ind, int step = 1) const
    {
	return parent->inc_index(ind, step);
    }
};

class Generic_Num_Subarray
    : public _Indirect_Numarray
{
    index_type origin[MAX_DIMENSION];
    index_type steps[MAX_DIMENSION];
    index_type the_end[MAX_DIMENSION];
    void init(const index_type *the_shape,
	      const index_type *the_origin,
	      const index_type *the_steps) 
    {
	int i;
	
	if(the_origin)
	    memcpy(origin, the_origin, MAX_DIMENSION * sizeof(index_type));
	else
	    memset(origin, 0, MAX_DIMENSION * sizeof(index_type));
	
	if(the_steps)
	    memcpy(steps, the_steps, MAX_DIMENSION * sizeof(index_type));
	else
	    for(i=0; i<MAX_DIMENSION; i++)
		steps[i] = 1;
	
	if(the_shape)
	    for(i=0; i<MAX_DIMENSION; i++)
		the_end[i] = origin[i] + the_shape[i];
	else
	    for(i=0; i<MAX_DIMENSION; i++)
		the_end[i] = origin[i] + parent->shape_ref(i);
    }

public:
    Generic_Num_Subarray() 
    {;}
    Generic_Num_Subarray(_Num_Array& the_parent,
			 const index_type *the_shape =NULL,
			 const index_type *the_origin =NULL,
			 const index_type *the_steps =NULL)
    {
	parent = &the_parent;
	init(the_shape, the_origin, the_steps);
    }
    ~Generic_Num_Subarray() {}
    void reshape(dim_type new_dimension, const index_type *new_shape)
    {
	init(new_shape, origin, steps);
    }
    void permute(const dim_type *new_order) // not really a valid function
    {
	parent->permute(new_order);
    }

    _ND_Array& block_ref(const index_type *the_shape,
			 const index_type *the_origin = NULL,
			 const index_type *the_steps = NULL) const
    {
	index_type actual_origin[MAX_DIMENSION];
	index_type actual_steps[MAX_DIMENSION];
	for(int i=0; i<MAX_DIMENSION; i++) {
	    actual_origin[i] = origin[i] + the_origin[i];
	    actual_steps[i]  = steps[i]  * the_steps[i];
	}
	return parent->block_ref(the_shape, actual_origin, actual_steps);
    }
    void block_set(const _ND_Array& new_value,
		   const index_type *the_origin = NULL,
		   const index_type *the_steps  = NULL)
    {
	index_type actual_origin[MAX_DIMENSION];
	index_type actual_steps[MAX_DIMENSION];
	for(int i=0; i<MAX_DIMENSION; i++) {
	    actual_origin[i] = origin[i] + the_origin[i];
	    actual_steps[i]  = steps[i]  * the_steps[i];
	}
	parent->block_set(new_value, actual_origin, actual_steps);
    }
    ND_Subarray* block_iref(const index_type *the_shape,
			    const index_type *the_origin = NULL,
			    const index_type *the_steps = NULL) const
    {
	index_type actual_origin[MAX_DIMENSION];
	index_type actual_steps[MAX_DIMENSION];
	for(int i=0; i<MAX_DIMENSION; i++) {
	    actual_origin[i] = origin[i] + the_origin[i];
	    actual_steps[i]  = steps[i]  * the_steps[i];
	}
	return parent->block_iref(the_shape, actual_origin, actual_steps);
    }
    
    _Num_Array& Num_block_ref(const index_type *the_shape,
			      const index_type *the_origin = NULL,
			      const index_type *the_steps = NULL) const
    {
	index_type actual_origin[MAX_DIMENSION];
	index_type actual_steps[MAX_DIMENSION];
	for(int i=0; i<MAX_DIMENSION; i++) {
	    actual_origin[i] = origin[i] + the_origin[i];
	    actual_steps[i]  = steps[i]  * the_steps[i];
	}
	return parent->Num_block_ref(the_shape, actual_origin, actual_steps);
    }
    void Num_block_set(const _Num_Array& new_value,
		       const index_type *the_origin = NULL,
		       const index_type *the_steps = NULL)
    {
	index_type actual_origin[MAX_DIMENSION];
	index_type actual_steps[MAX_DIMENSION];
	for(int i=0; i<MAX_DIMENSION; i++) {
	    actual_origin[i] = origin[i] + the_origin[i];
	    actual_steps[i]  = steps[i]  * the_steps[i];
	}
	parent->Num_block_set(new_value, actual_origin, actual_steps);
    }
    Num_Subarray* Num_block_iref(const index_type *the_shape,
				 const index_type *the_origin = NULL,
				 const index_type *the_steps = NULL) const
    {
	index_type actual_origin[MAX_DIMENSION];
	index_type actual_steps[MAX_DIMENSION];
	for(int i=0; i<MAX_DIMENSION; i++) {
	    actual_origin[i] = origin[i] + the_origin[i];
	    actual_steps[i]  = steps[i]  * the_steps[i];
	}
	return parent->Num_block_iref(the_shape, actual_origin, actual_steps);
    }

    T elem_ref(const index_type *index) const {
	index_type actual_index[MAX_DIMENSION];
	for(int i=0; i<MAX_DIMENSION; i++)
	    actual_index[i] = steps[i] * index[i] + origin[i];
	return parent->elem_ref(actual_index);
    }

    void elem_set(const index_type *index, T new_value) {
	index_type actual_index[MAX_DIMENSION];
	for(int i=0; i<MAX_DIMENSION; i++)
	    actual_index[i] = steps[i] * index[i] + origin[i];
	parent->elem_set(actual_index, new_value);
    }

    T* elem_iref(const index_type *index) const {
	index_type actual_index[MAX_DIMENSION];
	for(int i=0; i<MAX_DIMENSION; i++)
	    actual_index[i] = steps[i] * index[i] + origin[i];
	return parent->elem_iref(actual_index);
    }

};
    

class _Extended_Numarray
: public _Indirect_Numarray
{
public:
//     _Extended_Numarray() {;}
//     _Extended_Numarray(const _Num_Array& parent) {;}
//     ~_Extended_Numarray() {;}

    void reshape(dim_type new_dimension, const index_type *new_shape)
    {
	parent->reshape(new_dimension, new_shape);
    }
    void permute(const dim_type *new_order)
    {
	parent->permute(new_order);
    }

    _ND_Array& block_ref(const index_type *shape,
			 const index_type *origin = NULL,
			 const index_type *steps = NULL) const
    {
	return parent->block_ref(shape, origin, steps);
    }
    void block_set(const _ND_Array& new_value,
		   const index_type *origin = NULL,
		   const index_type *steps  = NULL)
    {
	parent->block_set(new_value, origin, steps);
    }
    ND_Subarray* block_iref(const index_type *shape,
			    const index_type *origin = NULL,
			    const index_type *steps = NULL) const
    {
	return parent->block_iref(shape, origin, steps);
    }
    
    _Num_Array& Num_block_ref(const index_type *shape,
			      const index_type *origin = NULL,
			      const index_type *steps = NULL) const
    {
	return parent->Num_block_ref(shape, origin, steps);
    }
    void Num_block_set(const _Num_Array& new_value,
		       const index_type *origin = NULL,
		       const index_type *steps = NULL)
    {
	parent->Num_block_set(new_value, origin, steps);
    }
    Num_Subarray* Num_block_iref(const index_type *shape,
				 const index_type *origin = NULL,
				 const index_type *steps = NULL) const
    {
	return parent->Num_block_iref(shape, origin, steps);
    }
};

class Extended_Numarray
    : public Indirect_Numarray
{
public:
//     Extended_Numarray() {;}
//     Extended_Numarray(const Num_Array& parent) {;}
//     ~Extended_Numarray() {;}

    void reshape(dim_type new_dimension, const index_type *new_shape)
    {
	parent->reshape(new_dimension, new_shape);
    }
    void permute(const dim_type *new_order)
    {
	parent->permute(new_order);
    }

    _ND_Array& block_ref(const index_type *shape,
			 const index_type *origin = NULL,
			 const index_type *steps = NULL) const
    {
	return parent->block_ref(shape, origin, steps);
    }
    void block_set(const _ND_Array& new_value,
		   const index_type *origin = NULL,
		   const index_type *steps  = NULL)
    {
	parent->block_set(new_value, origin, steps);
    }
    ND_Subarray* block_iref(const index_type *shape,
			    const index_type *origin = NULL,
			    const index_type *steps = NULL) const
    {
	return parent->block_iref(shape, origin, steps);
    }
    
    _Num_Array& Num_block_ref(const index_type *shape,
			      const index_type *origin = NULL,
			      const index_type *steps = NULL) const
    {
	return parent->Num_block_ref(shape, origin, steps);
    }
    void Num_block_set(const _Num_Array& new_value,
		       const index_type *origin = NULL,
		       const index_type *steps = NULL)
    {
	parent->Num_block_set(new_value, origin, steps);
    }
    Num_Subarray* Num_block_iref(const index_type *shape,
				 const index_type *origin = NULL,
				 const index_type *steps = NULL) const
    {
	return parent->Num_block_iref(shape, origin, steps);
    }
};
    
//
// We need an arithmetically sensible modulo operator.  As it is to be
// used in calculating indices into arrays it also needs to be efficient:
//
inline int modulo(int n1, int n2) {
    int i = n1 % n2;
    return ( ((n1 < 0) != (n2 < 0) && (i != 0))  ? i+n2 : i );
}

class Padded__Numarray
: public _Extended_Numarray
{
  protected:
    T out_of_bounds_value;
  public:
    Padded__Numarray() : out_of_bounds_value(0) {;}
    Padded__Numarray(const _Num_Array& parent, T pad_value = 0)
    {
	parent_set(parent);
	out_of_bounds_value = pad_value;
    }

    virtual ~Padded__Numarray() {;}

    T elem_ref(const index_type *index) const {
       return (inside(index) ? parent->elem_ref(index) 
	       : out_of_bounds_value);
    }

    void elem_set(const index_type *index, T new_value) {
	if(inside(index)) 
	    elem_set(index, new_value);
	else
	    out_of_bounds_value = new_value;
    }

    T* elem_iref(const index_type *index) const {
	return (inside(index) ? parent->elem_iref(index) 
		: (T *)&out_of_bounds_value);
    }

};

class Padded_Numarray
: public Extended_Numarray
{
  protected:
    T out_of_bounds_value;
  public:
    Padded_Numarray() : out_of_bounds_value(0) {;}
    Padded_Numarray(const Num_Array& parent, T pad_value = 0)
    {
	parent_set(parent);
	out_of_bounds_value = pad_value;
    }

    virtual ~Padded_Numarray() {;}

    T elem_ref(const index_type *index) const {
       return (inside(index) ? parent->felem_ref(index) 
	       : out_of_bounds_value);
    }

    void elem_set(const index_type *index, T new_value) {
	if(inside(index)) 
	    parent->felem_set(index, new_value);
	else
	    out_of_bounds_value = new_value;
    }

    T* elem_iref(const index_type *index) const {
	return (inside(index) ? parent->felem_iref(index) 
		: (T *)&out_of_bounds_value);
    }

    /*
     * The following are the same as the above but are not vitual and therefore
     * may be inlined by the compiler to produce much faster code:
     */
    T felem_ref(const index_type *index) const {
       return (inside(index) ? parent->felem_ref(index) 
	       : out_of_bounds_value);
    }

    void felem_set(const index_type *index, T new_value) {
	if(inside(index)) 
	    parent->felem_set(index, new_value);
	else
	    out_of_bounds_value = new_value;
    }

    T* felem_iref(const index_type *index) const {
	return (inside(index) ? parent->felem_iref(index) 
		: (T *)&out_of_bounds_value);
    }
};

class Periodic__Numarray
    : public _Extended_Numarray
{
protected:
    index_type* periodise(const index_type *index) const {
	static index_type _ext_index[MAX_DIMENSION];
	for(dim_type i=0; i<dimension_ref(); i++)
	    _ext_index[i] =
		(inside(index,i) ? index[i] : modulo(index[i],shape_ref(i)));
	return(_ext_index);
    }
public:
    Periodic__Numarray() {;}
    Periodic__Numarray(const _Num_Array& parent)
    {
	parent_set(parent);
    }
    
    virtual ~Periodic__Numarray() {;}
    
    T elem_ref(const index_type *index) const {
	return parent->elem_ref(periodise(index));
    }
    void elem_set(const index_type *index, T new_value) {
	parent->elem_set(periodise(index), new_value);
    }
    T* elem_iref(const index_type *index) const {
	return parent->elem_iref(periodise(index));
    }
};

class Periodic_Numarray
    : public Extended_Numarray
{
protected:
    index_type* periodise(const index_type *index) const {
	static index_type _ext_index[MAX_DIMENSION];
	for(dim_type i=0; i<dimension_ref(); i++)
	    _ext_index[i] =
		(inside(index,i) ? index[i] : modulo(index[i],shape_ref(i)));
	return(_ext_index);
    }
public:
    Periodic_Numarray() {;}
    Periodic_Numarray(const Num_Array& parent)
    {
	parent_set(parent);
    }
    
    virtual ~Periodic_Numarray() {;}
    
    T elem_ref(const index_type *index) const {
	return parent->felem_ref(periodise(index));
    }
    void elem_set(const index_type *index, T new_value) {
	parent->felem_set(periodise(index), new_value);
    }
    T* elem_iref(const index_type *index) const {
	return parent->felem_iref(periodise(index));
    }

    T felem_ref(const index_type *index) const {
	return parent->felem_ref(periodise(index));
    }
    void felem_set(const index_type *index, T new_value) {
	parent->felem_set(periodise(index), new_value);
    }
    T* felem_iref(const index_type *index) const {
	return parent->felem_iref(periodise(index));
    }
};

#endif	// __IND_ARRAY__
