/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

#include "ND_Array.hh"

/*
 * Support functions
 */

index_type *
pack_shape(int step,
	   int dim,  const index_type *shape, 
	   int pdim,       index_type *pshape)
{
    dim_type i;

    while(dim > pdim)
	for(i=0, dim--, pdim--; i<step; i++, dim--)
	    pshape[pdim-i] = shape[dim] * shape[dim-step];
    for(pdim-=step; pdim>=0; pdim--)
	pshape[pdim] = shape[pdim];
    return pshape;
}

void 
_ND_Array::copy(const _ND_Array& other) 
{
    unsigned long size = size_ref();
    index_type ind[MAX_DIMENSION];
    
    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	elem_set(ind, other.elem_ref(ind));
	inc_index(ind);
    }
}

void 
_ND_Array::permutable_copy(const _ND_Array& other) 
{
    unsigned long size = size_ref();
    index_type ind1[MAX_DIMENSION];
    index_type ind2[MAX_DIMENSION];
    
    memset(ind1, 0, MAX_DIMENSION * sizeof(index_type));
    memset(ind2, 0, MAX_DIMENSION * sizeof(index_type));
    
    for(unsigned long i=0; i<size; i++) {
	elem_set(ind1, other.elem_ref(ind2));
	inc_index(ind1);
	other.inc_index(ind2);
    }
}

void
_ND_Array::rotate()
{
    dim_type new_order[MAX_DIMENSION];
    new_order[dimension_ref()-1] = 0;
    for(dim_type i=1; i<dimension_ref(); i++)
	new_order[i-1] = i;
    permute(new_order);
}

//
// ND_Array
//	Private member functions:

void 
ND_Array::init (dim_type new_dimension, 
		 const index_type *new_shape,
		 storage_convension new_convension)
{
    dimension = new_dimension;
    convension = new_convension;
    if(new_shape)
	memcpy(shape, new_shape, dimension * sizeof(index_type));
    switch(new_convension)
	{
	case C:
	    if(new_dimension > 0){
		dim_type end=new_dimension-1;
		
		if(new_shape) {
		    num_elements = new_shape[end];
		    multipliers[0] = 1;
		    for(int i=0; i<end; i++) {
			num_elements *= new_shape[i];
			multipliers[i+1] = multipliers[i] * new_shape[i];
		    }
		} else {
		    num_elements = 0;
		    for(int i=0; i<end; i++)
			multipliers[i] = 0;
		}
	    } else
		num_elements = 0;
	    break;
	case Fortran:
	    if(new_dimension > 0){
		dim_type end=new_dimension-1;
		
		if(new_shape) {
		    num_elements = 1;
		    multipliers[end] = 1;
		    for(int i=end; i>=0; i--) {
			num_elements *= new_shape[i];
			multipliers[i-1] = multipliers[i] * new_shape[i];
		    }
		} else {
		    num_elements = 0;
		    for(int i=0; i<end; i++)
			multipliers[i] = 0;
		}
	    } else
		num_elements = 0;
	    break;
	default:
	    fprintf(stderr, "bad convension in array initialisation %i\n",
		    convension);
	}
}

// index_type *
// ND_Array::inc_index(index_type *ind, int step) const
// {
//     ind[0]+=step;
//     if( ind[0] >= shape[0] ) {
// 	for( dim_type i=0; i<dimension-1; i++ ) {
// 	    if( ind[i] >= shape[i] ) {
// 		if( i != dimension-1 ) {
// 		    ind[i] = 0;
// 		    ind[i+1]+=step;
// 		}
// 	    }
// 	}
//     }
//     return ind;
// }

void 
ND_Array::reshape(dim_type new_dimension, const index_type *new_shape)
{
    /*
     * ND equivalent of the MATLAB reshape function:  Reads data into the new
     * array row-wise.  If the storage convension is native 'C' then this
     * simplifies to a re-initialization of the shape (and multipliers).
     */
    if(convension != C && convension != Fortran)
	convension_set(C);
    init(new_dimension, new_shape, convension);
}

void 
ND_Array::copy (const _ND_Array& other)
{
    T *sp = storage;
    unsigned long size = size_ref();
    index_type ind[MAX_DIMENSION];

    if(convension == C) {
	memset(ind, 0, other.dimension_ref() * sizeof(index_type));
	for(unsigned long l=0; l<size; l++) {
	    *sp = other.elem_ref(ind);
	    sp++;
	    other.inc_index(ind);
	}
    }
    else
	_ND_Array::copy(other);
}

void
ND_Array::copy (const ND_Array& other)
{
    if((convension == other.convension_ref())
       && (convension != Unknown))
	switch(allocation)
	    {
	    case New:
	    case Static:
		for(unsigned int i=0; i<num_elements; i++)
		    storage[i] = other.storage[i];
		break;
	    case Malloc:
		memcpy(storage, other.storage_iref(), size_ref() * sizeof(T));
		break;
	    default:
		fprintf(stderr, "Unknown allocation in copy %d\n", allocation);
	    }
    else if(convension == C)
	ND_Array::copy(* (_ND_Array *) (&other));
    else
	_ND_Array::copy(other);
}

void
ND_Array::permute (const dim_type *new_order)
{
    index_type new_shape[MAX_DIMENSION];
    index_type old_multipliers[MAX_DIMENSION];
    dim_type i;

    for(i=0; i<dimension; i++)
	new_shape[i] = shape_ref(new_order[i]);
    memcpy(shape, new_shape, dimension * sizeof(index_type));

    memcpy(old_multipliers, multipliers, dimension*sizeof(index_type));
    for(i=0; i<dimension; i++)
	multipliers[i] = old_multipliers[new_order[i]];
    convension = Unknown;
}

void 
ND_Array::shallow_copy(ND_Array& other)
{
    dimension = other.dimension;
    memcpy(shape, other.shape, dimension * sizeof(index_type));
    convension = other.convension;
    memcpy(multipliers, other.multipliers, dimension * sizeof(index_type));
    num_elements = other.num_elements;

    switch(allocation) {
    case Malloc:	free(storage);		break;
    case New:		delete [] storage;	break;
    case Static:				break;
    }

    switch(other.allocation) {
    case Malloc:
    case New:		
	storage = other.storage;	
	break;
    case Static:
	storage = short_storage;
	for(unsigned long i=0; i<num_elements; i++)
	    storage[i] = other.storage[i];
	break;
    }
    allocation = Static;	// we are not responsible for deallocation    
}
    

//
//	Constructors and Destructors:
 
ND_Array::ND_Array (dim_type the_dimension, 
		    const index_type *the_shape,
		    storage_convension the_convension,
		    storage_allocation the_allocation,
		    T *the_storage)
    : num_elements(0), 
      storage(0)
{
    init(the_dimension, the_shape, the_convension);
    if(the_storage==NULL) {
      if(num_elements > SHORT_STORAGE_LENGTH) {
	switch(the_allocation) {
	case New:
	  storage = new T[num_elements];
	  allocation = New;
	  break;
	case Malloc:
	  storage = (T *)malloc(sizeof(T) * num_elements);
	  allocation = Malloc;
	  break;
	case Static: 	/* this is illegal---will never happen	*/
	  fprintf(stderr,
		  "static allocation error in ND_Array::ND_Array");
	  break;
	}
      }
      else {
	storage = short_storage;
	allocation = Static;
      }
    }
    else {
      allocation = the_allocation;
      storage = the_storage;
    }
}
 
ND_Array::ND_Array(const _ND_Array& other, 
		   storage_convension the_convension,
		   storage_allocation the_allocation)
{
    init(other.dimension_ref(), 
	 other.shape_ref(), 
	 the_convension); 

    switch(the_allocation)
	{
	case New:
	    storage = new T[num_elements];
	    allocation = New;
	    break;
	case Malloc:
	    storage = (T *)malloc(sizeof(T) * num_elements);
	    allocation = Malloc;
	    break;
	case Static:	    /* Not handled but should be-can lead to errors */
	    fprintf(stderr,
			"static allocation error in ND_Array::ND_Array");
	    break;
	}  

    copy(other);
}

ND_Array::ND_Array(const ND_Array& other, 
		   storage_convension the_convension,
		   storage_allocation the_allocation)
{
    init(other.dimension_ref(), 
	 other.shape_ref(), 
	 the_convension); 

    switch(the_allocation)
	{
	case New:
	    storage = new T[num_elements];
	    allocation = New;
	    break;
	case Malloc:
	    storage = (T *)malloc(sizeof(T) * num_elements);
	    allocation = Malloc;
	    break;
	case Static:	    /* Not handled but should be-can lead to errors */
	    fprintf(stderr,
			"static allocation error in ND_Array::ND_Array");
	    break;
	}  

    copy(other);
}

ND_Array::~ND_Array()
{
    switch(allocation)
	{
	case Malloc:	free(storage);		break;
	case New:	delete [] storage;	break;
	case Static:	/* do nothing */	break;
	}
}

//
//	Member functions for the abstract class:

static index_type *
_index_reverse(index_type *ind1, index_type *ind2, int length)
{
    ind2 = ind2 + length - 1;
    while(length-- > 0) {
	*ind2++ = *ind1--;
    }
    return ind2;
}

void 
ND_Array::convension_set(storage_convension new_convension)
{
    ND_Array *tmp;

    if(new_convension != convension) {
	/*
	 * Make a new ND Array with the required convension and copy into it.
	 */
	tmp = new ND_Array(dimension, shape, new_convension, allocation);
	tmp->copy(*this);
	/*
	 * Return the result to 'this' using a shallow copy then fix the
	 * allocation so that tmp will not deallocate storage on destruction
	 */
	this->shallow_copy(*tmp);
	allocation = tmp->allocation;
	tmp->allocation = Static;	// no deallocation will be done in tmp
	delete tmp;
    }
}

T* 
ND_Array::storage_ref(storage_convension the_convension) const
{
    /*
     * the_allocation (New, Malloc) --- static would require an output arg
     */
    T* sptr = NULL;
    unsigned long i;
    ND_Array *tmp = NULL;

    if(convension != the_convension) {
	tmp = new ND_Array(dimension, shape, the_convension, allocation);
	tmp->copy(*this);
	sptr = tmp->storage_iref();
    }
    else
	sptr = storage;

    T *copy_of_data = NULL;
    switch(allocation)
	{
	case New:	
	    copy_of_data = new T[num_elements];
	    for(i=0; i<num_elements; i++)
		copy_of_data[i] = sptr[i];
	    break;
	case Static:	/* Assume we can do a binary copy	*/
	case Malloc:	
	    copy_of_data = (T*)malloc(num_elements * sizeof(T));
	    memcpy(copy_of_data, sptr, num_elements * sizeof(T));
	    break;
	}

    if(tmp != NULL)	delete tmp;

    return copy_of_data;
}

void 
ND_Array::storage_set(T* new_storage, 
		      storage_convension new_convension,
		      storage_allocation new_allocation) 
{
    if(convension == Unknown)
	convension_set(C);

    switch(allocation)
	{
	case Malloc:	free(storage);		break;
	case New:	delete [] storage;	break;
	case Static:				break;
	}

    storage = new_storage;
    allocation = new_allocation;
    if(convension != new_convension) {
	convension = new_convension;
	/*
	 * Now reverse the multipliers to account for reverse ordering
	 * of array data in Fortran Vs. C
	 * This needs to be made more general (urgently).
	 */
	index_type tmp[MAX_DIMENSION];
	memcpy(tmp, multipliers, dimension * sizeof(index_type));
	_index_reverse(tmp, multipliers, dimension);
    }
}
 
T 
ND_Array::elem_ref(const index_type *index) const
{
    unsigned long storage_address = index[0] * multipliers[0];
    for(dim_type i=1; i<dimension; i++)
	storage_address += index[i] * multipliers[i];
    return storage[storage_address];
}

void 
ND_Array::elem_set(const index_type *index, T new_value)
{
    unsigned long storage_address = index[0] * multipliers[0];
    for(dim_type i=1; i<dimension; i++)
	storage_address += index[i] * multipliers[i];
    storage[storage_address] = new_value;
}

 
T*
ND_Array::elem_iref(const index_type *index) const
{
    unsigned long storage_address = index[0] * multipliers[0];
    for(dim_type i=1; i<dimension; i++)
	storage_address += index[i] * multipliers[i];
    return &storage[storage_address];
}

_ND_Array& 
ND_Array::block_ref(const index_type *shape,
		    const index_type *origin,
		    const index_type *steps) const
{
    ND_Subarray iref(*this, shape, origin, steps);
    ND_Array	*block = new ND_Array(iref.dimension_ref(), iref.shape_ref());
    block->copy(iref);
    return *block;
}
    
void 
ND_Array::block_set(const _ND_Array& new_value,
		    const index_type *origin,
		    const index_type *steps)
{
    ND_Subarray iref(*this, new_value.shape_ref(), origin, steps);
    iref.copy(new_value);
}

ND_Subarray * 
ND_Array::block_iref(const index_type *shape,
		     const index_type *origin,
		     const index_type *steps) const
{
    return (new ND_Subarray(*this, shape, origin, steps));
}

//
//	Additional member functions:
// These have been axed to avoid linking with the streamio class libraries
// void 
// ND_Array::print() const
// {
//     dim_type dim = dimension_ref();
//     index_type ind[MAX_DIMENSION];
//     dim_type i;

//     memset(ind,0,dim*sizeof(index_type));
//     while(inside(ind, dim-1)) {
// 	cout << "Row starting at [";
// 	for(i=0; i<dim; i++)
// 	    cout << ind[i] << ",";
// 	cout << "]\n";
//         while(inside(ind,0)) {
//             cout << "\t" << elem_ref(ind);
//             ind[0]++;
//         }
// 	cout << "\n";
//         for(i=0; i<dim; i++) {
//             if(!inside(ind, i)) {
//                 if(i!=dim-1) {
//                     ind[i] = 0;
//                     ind[i+1]++;
//                 }
//             }
//         }
//     }
// }

// void 
// ND_Array::_dump() const
// {
//     unsigned long nelem = size_ref();

//     for(unsigned long i=0; i<nelem; i++)
// 	cout << storage[i] << " ";
// }



//  SUBARRAY
//
// Implementation:
//	Constructors and destructors:
//
void 
ND_Subarray::init(dim_type the_dimension, 
		  const index_type *the_shape,
		  const index_type *the_origin,
		  const index_type *the_steps)
{
    dimension = the_dimension;
    if(the_shape)
	memcpy(shape, the_shape, dimension * sizeof(index_type));
    if(the_dimension >0) {
	if(the_shape) {
	    index_type *shapep=shape, *multp=multipliers;
	    offset = 0;
	    num_elements = 1;
	    for(dim_type i=0; i<parent->dimension; i++) {
		origin[i]  = (the_origin ? the_origin[i] : 0);
		steps[i]   = (the_steps  ?  the_steps[i] : 1);
		the_end[i] = (the_origin ? the_origin[i] : 0) + 
			     (the_steps  ?  the_steps[i] : 1) *
			     (the_shape[i] - 1); 
		// if(the_shape[i] > 1) { // && the_steps[i] != 0
		num_elements *= the_shape[i];
		*shapep = the_shape[i]; 
		shapep++;
		*multp = parent->multipliers[i] *
		  (the_steps ? the_steps[i] : 1);
		multp++;
		//}
		offset += parent->multipliers[i] * 
		  (the_origin ? the_origin[i] : 0);
	    }
	}
    }
}

void 
ND_Subarray::move(const index_type *inc)
{
    index_type i;
    /*
     * Moves the position of a subarray by the amount specified in `inc'
     * ie.  THIS IS A RELATIVE MOVE---NOT AN ABSOLUTE MOVE
     */
    for(i=0; i<parent->dimension; i++){
	if(inc[i]){
	    if(steps[i])
		offset += parent->multipliers[i] * inc[i];
	    origin[i]  += inc[i];
	    the_end[i] += inc[i];
	}
    }
}

void 
ND_Subarray::inc_position(const index_type *inc)
{
    index_type i, ind[MAX_DIMENSION], new_end[MAX_DIMENSION];

    memset(ind, 0, parent->dimension*sizeof(index_type));
    memcpy(new_end, the_end, parent->dimension*sizeof(index_type));
    /*
     * The following code finds the first dimension (in stored order) which,
     * when incremented by the amount specified in `inc' (or 1 if inc is null)
     * produces a subarray which is still `inside' its parent, and performs
     * the increment.  The function is used to "shuffle" a subarray through
     * its parent in the convensional mannor (left->right; top->bottom; etc).
     */
    for(i=0; i<parent->dimension; i++) {
	if(inc ? inc[i] : 1) {
	    new_end[i] += (inc ? inc[i] : 1);
	    if(parent->inside(new_end, i)) {
		ind[i] += (inc ? inc[i] : 1);
		move(ind);
		return;
	    }
	    else {
		ind[i] -= origin[i];	// = -origin[i] as ind[i]==0 here
		new_end[i] = 0;		// This value just has to be inside
	    }
	}
    }
    move(ind);
}

void
ND_Subarray::permute (const dim_type *new_order)
{
    index_type new_shape[MAX_DIMENSION];
    index_type old_multipliers[MAX_DIMENSION];
    dim_type i;

    for(i=0; i<dimension; i++)
	new_shape[i] = shape_ref(new_order[i]);
    memcpy(shape, new_shape, dimension * sizeof(index_type));

    memcpy(old_multipliers, multipliers, dimension*sizeof(index_type));
    for(i=0; i<dimension; i++)
	multipliers[i] = old_multipliers[new_order[i]];
}

void 
ND_Subarray::reshape(dim_type new_dimension, const index_type *new_shape)
{
    init(new_dimension, new_shape, NULL, NULL);
}

void 
ND_Subarray::reshape(dim_type new_dimension, 
		    const index_type *new_shape,
		    const index_type *new_origin,
		    const index_type *new_steps)
{
    init(new_dimension, new_shape, new_origin, new_steps);
}

ND_Subarray::ND_Subarray(const ND_Array& the_parent,
                const index_type *the_shape,
                const index_type *the_origin,
                const index_type *the_steps)
    : parent(0),
      offset(0), 
      num_elements(0)
{
//    dim_type dim = 0;

    parent_set(the_parent);
//     if(the_shape)
// 	for(dim_type i=0; i<the_parent.dimension_ref(); i++)
// 	    if(the_shape[i] > 1 || 
// 	       (the_shape[i]==1 && the_parent.shape_ref(i)==1))
// 		dim++;
    init(the_parent.dimension_ref(), the_shape, the_origin, the_steps);
}

ND_Subarray::ND_Subarray(const ND_Subarray& the_parent,
                const index_type *the_shape,
                const index_type *the_origin,
                const index_type *the_steps)
    : parent(0),
      offset(0), 
      num_elements(0)
{
//    dim_type dim = 0;

    parent_set(*the_parent.parent);
//     if(the_shape)
// 	for(dim_type i=0; i<the_parent.dimension_ref(); i++)
// 	    if(the_shape[i] > 1 || 
// 	       (the_shape[i]==1 && the_parent.shape_ref(i)==1))
// 		dim++;
    init(the_parent.dimension_ref(), the_shape, the_origin, the_steps);
}

 T 
ND_Subarray::elem_ref(const index_type *index) const
{
    unsigned long storage_address = index[0]*multipliers[0];
    for(dim_type i=1; i<dimension; i++)
        storage_address += index[i]*multipliers[i];
    return parent->storage[storage_address+offset];
}

void 
ND_Subarray::elem_set(const index_type *index, T new_value)
{
    unsigned long storage_address = index[0]*multipliers[0];
    for(dim_type i=1; i<dimension; i++)
        storage_address += index[i]*multipliers[i];
    parent->storage[storage_address+offset] = new_value;
}

 T* 
ND_Subarray::elem_iref(const index_type *index) const
{
    unsigned long storage_address = index[0]*multipliers[0];
    for(dim_type i=1; i<dimension; i++)
        storage_address += index[i]*multipliers[i];
    return &(parent->storage[storage_address+offset]);
}

_ND_Array& 
ND_Subarray::block_ref(const index_type *shape,
		       const index_type *origin,
		       const index_type *steps) const
{
    ND_Subarray iref(*this, shape, origin, steps);
    ND_Array	*block = new ND_Array(iref.dimension_ref(), iref.shape_ref());
    block->copy(iref);
    return *block;
}
    
void 
ND_Subarray::block_set(const _ND_Array& new_value,
		       const index_type *origin,
		       const index_type *steps)
{
    ND_Subarray iref(*this, new_value.shape_ref(), origin, steps);
    iref.copy(new_value);
}

ND_Subarray * 
ND_Subarray::block_iref(const index_type *shape,
			const index_type *origin,
			const index_type *steps) const
{
    return (new ND_Subarray(*this, shape, origin, steps));
}

