/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

#ifndef __ND_Array__
#define __ND_Array__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream.h>

typedef unsigned char dim_type;
const MAX_DIMENSION = 64;
typedef long index_type;
enum storage_convension {C, Fortran, Unknown};
enum storage_allocation {New, Malloc, Static};
typedef float Real;
typedef Real T;
// typedef int bool;
// #define true 1
// #define false 0
//----------------------------------------------------------------------------
// The first part here defines some abstract classes for N-Dimensional
//	arrays and two useful specialisations---Matrices and Vectors.
//	Abstract classes are defined so that a number of different
//	implementations can be realised in a "nice" way.  In particular
//	subarrays (which reference a sub-region an array and may have smaller
//	dimension) are still treated as arrays as they inherit from the
//	same abstract class.  Similar logic applies for extended arrays.
//

class _ND_Array;
class ND_Array;
class ND_Subarray;

/*
 * _ND_Array:	The abstract class for N-Dimensional arrays
 */
class _ND_Array {
public:
    virtual dim_type dimension_ref() const = 0;
    
    virtual const index_type* shape_ref() const = 0;

    virtual index_type shape_ref(dim_type d) const { return shape_ref()[d]; }
    
    virtual unsigned long size_ref() const {
	unsigned long size = shape_ref(0);
	for(dim_type i=1; i <dimension_ref(); i++)
	    size *= shape_ref(i);
	return size;
    }

    bool inside(const index_type *the_index) const {
        bool in = (the_index[0] >= 0) && (the_index[0] < shape_ref(0));
        for(dim_type i=1; i<dimension_ref(); i++)
            in = in && (the_index[i] >= 0) && (the_index[i] < shape_ref(i));
        return in;
    }

    bool inside(const index_type *the_index, dim_type the_dimension) const {
        return (the_index[the_dimension] < shape_ref(the_dimension)) &&
            (the_index[the_dimension] >= 0);
    }

    bool conforms(const _ND_Array& other) const {
	bool same = (dimension_ref() == other.dimension_ref());
	for(dim_type i=0; i<dimension_ref(); i++)
	    same &= (shape_ref(i) == other.shape_ref(i));
	return same;
    }

    /*
     * This should be done better --- it cannot be fully inlined in its
     * present form... (as it calls puore virtual functions).
     */
    index_type *inc_index(index_type *ind, int step = 1) const
    {
	ind[0]+=step;
	if( ind[0] >= shape_ref(0) ) {
	    dim_type d = dimension_ref()-1;
	    for( dim_type i=0; i<d; i++ ) {
		if( ind[i] >= shape_ref(i) ) {
		    if( i != d ) {
			ind[i] = 0;
			ind[i+1]+=step;
		    }
		}
	    }
	}
	return ind;
    };
    
    /*
     * Pure virtual functions:
     *
     * NB. The element referencing and setting functions are useful when 
     *  generating highly reusable code.  For most applications however, they
     *	lead to an unacceptable overhead as each element reference or set
     *	necessitates a virtual function call!  To overcome this, a derived
     *	class may write its own fast (inlinable) reference functions, 
     *  prepending the standard reference and set functions with an 'f'
     *	(eg felem_ref(...))... Oh well
     */
    virtual T elem_ref(const index_type *index) const = 0;
    virtual void elem_set(const index_type *index, T new_value) = 0;
    virtual T *elem_iref(const index_type *index) const = 0;
    virtual _ND_Array& block_ref(const index_type *shape,
				 const index_type *origin = NULL,
				 const index_type *steps = NULL
	) const =0;
    virtual void block_set(const _ND_Array& new_value,
			   const index_type *origin = NULL,
			   const index_type *steps  = NULL
	) = 0;
    virtual ND_Subarray* block_iref(const index_type *shape,
				    const index_type *origin = NULL,
				    const index_type *steps = NULL
	) const = 0;
    virtual void reshape(dim_type new_dimension, 
			 const index_type *new_shape) = 0;
    virtual void permute(const dim_type *new_order) = 0;
    /*
     * Other virtual functions with definitions on the abstract class
     */
    virtual void copy(const _ND_Array& other);
    virtual void permutable_copy(const _ND_Array& other);
    virtual void rotate();
};

/*
 * Prototypes for functions associated with the _ND_Array class (but not
 * directly part of it).
 */
index_type *pack_shape(int step,
		       int dim,  const index_type *shape, 
		       int pdim,       index_type *pshape);



#define SHORT_STORAGE_LENGTH	256

//----------------------------------------------------------------------------
// Next, define the instantiable class ND_Array
//
class ND_Array 
    : public virtual _ND_Array
{
protected:
    dim_type dimension;
    index_type shape[MAX_DIMENSION];
    storage_convension convension;
    storage_allocation allocation;
    index_type multipliers[MAX_DIMENSION];
    unsigned long num_elements;
    T short_storage[SHORT_STORAGE_LENGTH];
    T *storage;
    void init (dim_type new_dimension, 
	       const index_type *new_shape,
	       storage_convension new_convension = C);
public:
    //
    // Constructors:
    //
    ND_Array(dim_type the_dimension = 0, 
	     const index_type *the_shape = NULL,
	     storage_convension new_convension = C,
	     storage_allocation the_allocation = New,
	     T *the_storage = NULL);
    ND_Array(const _ND_Array& other,
	     storage_convension new_convension = C,
	     storage_allocation the_allocation = New);
    ND_Array(const ND_Array& other,
	     storage_convension new_convension = C,
	     storage_allocation the_allocation = New);
    ~ND_Array();
    //
    // Member functions for the abstract class:
    //
    dim_type dimension_ref() const { return dimension; }
    const index_type *shape_ref() const { return shape; }
    index_type shape_ref(index_type d) const { return shape[d]; }
    unsigned long size_ref() const { return num_elements; }

    T elem_ref(const index_type *index) const;
    void elem_set(const index_type *index, T new_value);
    T *elem_iref(const index_type *index) const;

    T felem_ref(const index_type *index) const
    {
	unsigned long storage_address = index[0] * multipliers[0];
	for(dim_type i=1; i<dimension; i++)
	    storage_address += index[i] * multipliers[i];
	return storage[storage_address];
    };

    void felem_set(const index_type *index, T new_value)
    {
	unsigned long storage_address = index[0] * multipliers[0];
	for(dim_type i=1; i<dimension; i++)
	    storage_address += index[i] * multipliers[i];
	storage[storage_address] = new_value;
    };

    T *felem_iref(const index_type *index) const
    {
	unsigned long storage_address = index[0] * multipliers[0];
	for(dim_type i=1; i<dimension; i++)
	    storage_address += index[i] * multipliers[i];
	return &storage[storage_address];
    };

    _ND_Array& block_ref(const index_type *shape,
			const index_type *origin = NULL,
			const index_type *steps = NULL) const;
    void block_set(const _ND_Array& new_value,
		   const index_type *origin = NULL,
		   const index_type *steps = NULL);
    ND_Subarray* block_iref(const index_type *shape,
			    const index_type *origin = NULL,
			    const index_type *steps = NULL) const;

    void reshape(dim_type new_dimension, const index_type *new_shape);
    void copy(const _ND_Array& other);
    void permute(const dim_type *new_order);
    //
    // Member function specific to this implementation:
    //
    void copy(const ND_Array& other);	// full copy---new copy of data made
    void shallow_copy(ND_Array& other); // reference to data copied only
    storage_convension convension_ref() const {return convension;}
    void convension_set(storage_convension new_convension = C);
    T* storage_ref(storage_convension the_convension = C) const;
    void storage_set(T* new_storage, 
		     storage_convension new_convension = C,
	             storage_allocation new_allocation = New);
    T* storage_iref() const {return storage;}
    //
    // for debugging; do not depend upon
    //
    void print() const;
    void _dump() const;
    friend class ND_Subarray;	// I realise this may be poor design!
};

//----------------------------------------------------------------------------
// Subarray:
//	Classes defining arrays which are references to parts of other arrays.
//	In many ways these are are dangerous class as they provide an ability
//	to produce side-effects on an instance of a different class.  They
//	are however an extremely powerful and efficient mechanism for
//	implementing many array manipulations... just use with caution.
//

class ND_Subarray
    : public virtual _ND_Array
{
protected:
    dim_type dimension;
    index_type shape[MAX_DIMENSION];
    const ND_Array *parent;
    unsigned long offset;
    unsigned long num_elements;
    index_type multipliers[MAX_DIMENSION];
    index_type origin[MAX_DIMENSION];
    index_type steps[MAX_DIMENSION];
    index_type the_end[MAX_DIMENSION];
    void init(dim_type the_dimension, 
	      const index_type *the_shape,
	      const index_type *the_origin,
	      const index_type *the_steps);
    void parent_set(const ND_Array& the_parent) {
	parent = &the_parent;
    }
public:
    ND_Subarray() 
	: parent(0),
	  offset(0), 
	  num_elements(0)
    {}
    ND_Subarray(const ND_Array& the_parent,
                const index_type *the_shape =NULL,
                const index_type *the_origin =NULL,
                const index_type *the_steps =NULL);
    ND_Subarray(const ND_Subarray& the_parent,
                const index_type *the_shape =NULL,
                const index_type *the_origin =NULL,
                const index_type *the_steps =NULL);
    ~ND_Subarray() {}

    dim_type dimension_ref() const { return dimension; }
    const index_type *shape_ref() const { return shape; }
    index_type shape_ref(index_type d) const { return shape[d]; }
    unsigned long size_ref() const { return num_elements; }

    /*
     * Storage ref and set should be beefed up to return / set requested
     * raw arrays as copies of the subarray data.  This would require that
     * the calling function "cleans up" the copy afterward.  For this to be
     * done, an allocation field is desirable
     */
//     T *storage_ref(storage_convension the_convension = C);
//     void storage_set(T *new_storage, 
// 		     storage_convension the_convension = C);

    /*
     * References/sets to subarray data are implemented as remappings to the
     * parent data (which is not copied and is therefore prone to
     * side-effects).  The possibility of side-effects allows efficient
     * manipulation of subregions in multidimensional arrays --- ie
     * without needing to copy into tmp storage and copy back in after.
     */
    T elem_ref(const index_type *index) const;
    void elem_set(const index_type *index, T new_value);
    T *elem_iref(const index_type *index) const;
    _ND_Array& block_ref(const index_type *shape,
			 const index_type *origin = NULL,
			 const index_type *steps = NULL) const;
    void block_set(const _ND_Array& new_value,
		   const index_type *origin = NULL,
		   const index_type *steps  = NULL);
    ND_Subarray* block_iref(const index_type *shape,
			    const index_type *origin = NULL,
			    const index_type *steps = NULL) const;
    void reshape(dim_type new_dimension, const index_type *new_shape);
    void reshape(dim_type new_dimension, const index_type *new_shape,
		 const index_type *new_origin,
		 const index_type *new_steps  = NULL);
    void permute (const dim_type *new_order);
    /*
     * Some courtesy routines used when a subarray is moved through a
     * parent array, usually with some processing applied to each region.
     */
    bool inside() const {	// subarray is completely inside parent
	return( parent->inside(origin) && parent->inside(the_end) );
    }
    bool outside() const {	// subarray is completely outside parent
	return( !parent->inside(origin) && !parent->inside(the_end) );
    }
    void move(const index_type *inc);
    void inc_position(const index_type *inc = NULL);
};



#endif



