/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

#ifndef __Num_Array__
#define __Num_Array__

#include <ND_Array.hh>
#include <convolve.hh>
#include <stats.hh>
#include <math.h>

enum threshold_type {Binary, Magnitude, Hard, Soft};

class _Num_Array;
class Num_Array;
class Num_Subarray;

//
// We would like to use the abstract N-Dimensional Array class as the basis
//	for a hierachy of numeric array classes.  These classes have
//	additional operations defined on them (arethmetic ones) which do
//	not make sense for general arrays.

//-----------------------------------------------------------------------------
// _Num_Array: the abstract class for numeric arrays
//
class _Num_Array 
    : public virtual _ND_Array	// The abstract attributes of N-D Arrays
{
public:
    _Num_Array(void) {;}
    virtual ~_Num_Array() {;}
  void _Num_Array::copy(const _Num_Array& other) {
      _ND_Array::copy(other);
  };
    virtual _Num_Array& Num_block_ref(const index_type *shape,
				      const index_type *origin = NULL,
				      const index_type *steps = NULL
	) const = 0;
    virtual void Num_block_set(const _Num_Array& new_value,
			       const index_type *origin = NULL,
			       const index_type *steps = NULL) = 0;
    virtual Num_Subarray* Num_block_iref(const index_type *shape,
					 const index_type *origin = NULL,
					 const index_type *steps = NULL
	) const = 0;

    virtual _Num_Array& operator*=(T scalar);
    virtual _Num_Array& operator*=(const _Num_Array& other);
    virtual _Num_Array& operator+=(T scalar);
    virtual _Num_Array& operator+=(const _Num_Array& other);
    virtual _Num_Array& operator-=(T scalar);
    virtual _Num_Array& operator-=(const _Num_Array& other);
    virtual _Num_Array& operator/=(T scalar);
    virtual Num_Array* convolve_rows(const _Num_Array& kernel,
				     int step,
				     extension_type extension = Truncated,
				     int aux_arg = 0) const;
    virtual Num_Array* upsample_convolve_rows(const _Num_Array& kernel,
					      int step,
					      extension_type extension=Truncated,
					      int aux_arg = 0) const;
    virtual Num_Array* convolve(const _Num_Array& kernel,
				int step,
				extension_type extension = Truncated,
				int aux_arg = 0) const;
  //    virtual stat_type* statistics() const;
    virtual stat_type* stats_1(const _Num_Array *mask) const;
    virtual stat_type* stats_2(double mean, const _Num_Array *mask=NULL) const;
    virtual void fill(T scalar);
    virtual Real mean() const;
  //    virtual Real median() const;
    virtual Real max() const;
    virtual Real min() const;
    virtual void bound_upper(T upper_bound);
    virtual void bound_lower(T lower_bound);
    virtual void normalise(T lower_bound, T upper_bound);
    virtual void threshold(T threshold, threshold_type type = Binary,
			   T zero = 0,  T one = 1);
    virtual void threshold(const _Num_Array *threshold, 
			   threshold_type type = Binary,
			   T zero = 0,  T one = 1);
};

//-----------------------------------------------------------------------------
// Num_Array: An instantiable class of numeric arrays based on ND_Array
//
class Num_Array 
    : public virtual _Num_Array, 
      public ND_Array
{
public:

  Num_Array(dim_type the_dimension = 0, 
	    const index_type *the_shape = NULL,
	    storage_convension the_convension = C,
	    storage_allocation the_allocation = Malloc,
	    T *the_storage = NULL)
    : ND_Array(the_dimension, the_shape, 
			 the_convension, the_allocation, 
			 the_storage) {
  }
  
  Num_Array(const _Num_Array& other,
	    storage_convension new_convension = C,
	    storage_allocation the_allocation = Malloc)
    : ND_Array(other, new_convension, the_allocation) {
  }

  Num_Array(const Num_Array& other,
	    storage_convension new_convension = C,
	    storage_allocation the_allocation = Malloc)
    : ND_Array(other, new_convension, the_allocation) {
  }

  void Num_Array::copy(const Num_Array& other) {
    ND_Array::copy(other);
  };

    _Num_Array& Num_block_ref(const index_type *shape,
			      const index_type *origin = NULL,
			      const index_type *steps = NULL) const;
    void Num_block_set(const _Num_Array& new_value,
		       const index_type *origin = NULL,
		       const index_type *steps = NULL);
    Num_Subarray* Num_block_iref(const index_type *shape,
				 const index_type *origin = NULL,
				 const index_type *steps = NULL) const;

    Num_Array& operator*=(T scalar);
    Num_Array& operator*=(const Num_Array& other);

    Num_Array* convolve_rows(const _Num_Array& kernel,
			     int step,
			     extension_type extension = Truncated,
			     int aux_arg = 0) const;
    Num_Array* upsample_convolve_rows(const _Num_Array& kernel,
				      int step,
				      extension_type extension=Truncated,
				      int aux_arg = 0) const;
    Num_Array* convolve(const _Num_Array& kernel,
			int step,
			extension_type extension = Truncated,
			int aux_arg = 0) const;

    Num_Array *clone() const;

    void fill(T scalar);
    Real mean() const;
    Real sum() const;
    Real median() const;
    Real max() const;
    Real min() const;
    void bound_upper(T upper_bound);
    void bound_lower(T lower_bound);
    void normalise(T lower_bound, T upper_bound);
  void threshold(T threshold, threshold_type type = Binary,
		 T zero = 0,  T one = 1);
  void threshold(const Num_Array *threshold, threshold_type type = Binary,
		 T zero = 0,  T one = 1);
};

//-----------------------------------------------------------------------------
// Num_Subarray: An instantiable class of numeric subarrays 
//	based on ND_Subarray
//
class Num_Subarray
    :public virtual _Num_Array,
     public ND_Subarray
{
public:
    Num_Subarray() {;}
    Num_Subarray(const Num_Array& the_parent, 
		 const index_type *the_shape,
		 const index_type *the_origin =NULL,
		 const index_type *the_steps =NULL)
      : ND_Subarray(the_parent, the_shape, the_origin, the_steps) {;}

    Num_Subarray(const Num_Subarray& the_parent,
		 const index_type *the_shape =NULL,
		 const index_type *the_origin =NULL,
		 const index_type *the_steps =NULL)
      : ND_Subarray(the_parent, the_shape, the_origin, the_steps) {;}

    ~Num_Subarray() {;}

    _Num_Array& Num_block_ref(const index_type *shape,
			      const index_type *origin = NULL,
			      const index_type *steps = NULL) const;
    void Num_block_set(const _Num_Array& new_value,
		       const index_type *origin = NULL,
		       const index_type *steps = NULL);
    Num_Subarray* Num_block_iref(const index_type *shape,
				 const index_type *origin = NULL,
				 const index_type *steps = NULL) const;

};


#endif  /*__Num_Array__*/
