/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

/*
 * CONVOLVE.cc
 *
 * Implementation of the convolution function for abstract numeric arrays
 * (ie. class _Num_Array).
 */

#include <Num_Array.hh>
#include <Ind_Array.hh>
#include <convolve.hh>

/*
 * Utility classes
 */

/*
 * class cbuff
 *
 * This is a belated acknowledgement of the overhead involved in element
 * access in these classes.  The cbuff class is a simple circular buffer.
 * The push operation adds new values to the buffer.  Once the buffer is
 * full, calls to push will overwrite the oldest data in the buffer.  An
 * element may be accessed randomly using an index which is an offset from
 * the oldest value in the buffer.  The intended use of the class is to
 * buffer the window of data (extracted from a larger array) used in
 * convolution.
 */
class cbuff 
{
private:

    T *buffer;
    index_type size;
    index_type stored;
    index_type head;
    index_type tail;
    index_type modulo_inc(index_type variable)
    {
	return ((variable + 1) < size ? (variable + 1) : 0);
    }

public:

    cbuff(index_type the_size) 
    {
	buffer = new T[the_size];
	size = the_size;
	stored = head = tail = 0;
    }

    ~cbuff()
    {
	delete [] buffer;
    }

    void push(T new_item)
    {
	buffer[tail] = new_item;
	tail = modulo_inc(tail);
	if(stored == size)
	    head = modulo_inc(head);
	else
	    stored++;
    }

    T elem_ref(index_type i) const
    {
	return buffer[(i+head)%size];
    }
}; // class cbuff

/*
 * Utility Functions
 */

static offset_type 
periodicity_offset(symetry_type symetry)
{
    switch(symetry) {
    case wsw:
    case hsw:
	return Zero;
    case wsh:
    case hsh:
    case hah:
    case wah:
	return Half;
    case haw:
    case waw:
	return One;
    }
}

static offset_type 
symetry_offset(symetry_type symetry)
{
    switch(symetry) {
    case wsw:
    case wsh:
	return Zero;
    case hsw:
    case hsh:
    case hah:
    case haw:
	return Half;
    case wah:
    case waw:
	return One;
    }
}

static index_type
period_of_symetric_extension(index_type length,
			     symetry_type symetry)
{
    switch(symetry) {
    case wsw:	return (2 * length - 2);
    case wsh:
    case hsw:	return (2 * length - 1);
    case hsh:
    case hah:	return (2 * length);
    case wah:
    case haw:	return (2 * length + 1);
    case waw:	return (2 * length + 2);
    }
    return 0; //should never happen
}

static symetry_type
symetry_after_convolution(symetry_type input_symetry,
			  index_type filter_symetry)
{
    symetry_type result;

    switch(filter_symetry) {
    case wsh:	
	result = input_symetry;	break;
    case hsh:
	switch(input_symetry) {
	case wsw:	result = hsh;	break;
	case wsh:	result = hsw;	break;
	case hsw:	result = wsh;	break;
	case hsh:	result = wsw;	break;
	case hah:	result = waw;	break;
	case wah:	result = haw;	break;
	case haw:	result = wah;	break;
	case waw:	result = hah;	break;
	}
	break;
    case wah:
	switch(input_symetry) {
	case wsw:	result = waw;	break;
	case wsh:	result = wah;	break;
	case hsw:	result = haw;	break;
	case hsh:	result = hah;	break;
	case hah:	result = hsh;	break;
	case wah:	result = wsh;	break;
	case haw:	result = hsw;	break;
	case waw:	result = wsw;	break;
	}
	break;
    case hah:
	switch(input_symetry) {
	case wsw:	result = hah;	break;
	case wsh:	result = haw;	break;
	case hsw:	result = wah;	break;
	case hsh:	result = waw;	break;
	case hah:	result = wsw;	break;
	case wah:	result = hsw;	break;
	case haw:	result = wsh;	break;
	case waw:	result = hsh;	break;
	}
	break;
    }

    return result;
}

index_type 
period_length(index_type input_length, symetry_type input_symetry)
{
    return 1;
}

static symetry_type
symetry_after_downsampling(index_type length, 
			   symetry_type input_symetry, 
			   int step,
			   int downsample_offset)
{
    symetry_type	result;

    int both_different = period_length(length, input_symetry) % 2;
    int symetric = (input_symetry == wsw ||
		    input_symetry == wsh ||
		    input_symetry == hsw ||
		    input_symetry == hsh);
    if((downsample_offset == 0) || 
       (input_symetry     == waw && 
	downsample_offset == step-1))	// whole sample periodicity
	result = (symetric ? (both_different ? hsw : wsw)
		  : (both_different ? haw : waw));
    else	// half sample periodicity
	result = (symetric ? (both_different ? wsh : hsh)
		  : (both_different ? wah : waw));

    return result;
}

static void
determine_filter_symetries(const _Num_Array& filter, symetry_type *symetries)
{
    double	epsilon = 1e-12;	// a very small number
    dim_type	d;
    index_type	length, i, j;
    index_type  ind[MAX_DIMENSION];
    index_type  shape[MAX_DIMENSION];
    Num_Subarray *row;
    Real	c1, c2;
    _Num_Array* tmp = (_Num_Array *) &filter;

    memset(symetries, 0, filter.dimension_ref() * sizeof(symetry_type));
    for(d=0; d<filter.dimension_ref(); d++)
	shape[d] = 1;

    for(d=0; d<filter.dimension_ref(); d++) {
	shape[0] = filter.shape_ref(0);
	row = filter.Num_block_iref(shape);
	for(index_type r=0; r<filter.size_ref() / row->size_ref(); r++) {
	    for(i=0, j=length-1; i<length/2; i++, j--) {
		c1 = row->elem_ref(&i);
		c2 = row->elem_ref(&j);
		if((length = row->size_ref()) % 2) {// filter has odd length
		    if(fabs(c1) > epsilon && fabs(c2) > epsilon) {
			symetries[d] = ((c1 - c2) > epsilon ? wah : wsh);
			break;
		    }
		}
		else {	// filter has even length
		    if(fabs(c1) > epsilon && fabs(c2) > epsilon) {
			symetries[d] = ((c1 - c2) > epsilon ? hah : hsh);
			break;
		    }
		}
	    }
	    if(symetries[d] != 0)
		break;
	    row->inc_position();
	}
	delete row;
	tmp->rotate();
    }
}

static int
symetric_filter_group_delay(int filter_length,
			    symetry_type filter_symetry,
			    int step,
			    symetry_type extension)
{
    symetry_type sym_after_conv;
    int		 downsample_offset;
    int		 filter_delay;

    sym_after_conv = symetry_after_convolution(extension, filter_symetry);

    switch(periodicity_offset(sym_after_conv)) {
    case Zero:	downsample_offset = 0;		break;
    case Half:	downsample_offset = step / 2;	break;
    case One:	downsample_offset = step - 1;	break;
    }

    filter_delay = filter_length / 2;

    return (filter_delay - downsample_offset);
}

static void
determine_symetric_filter_group_delays(const _Num_Array& filter,
				       int *group_delays,
				       int step,
				       symetry_type extension)
{
    symetry_type symetries[MAX_DIMENSION];

    determine_filter_symetries(filter, symetries);
    for(dim_type d=0; d<filter.dimension_ref(); d++)
	group_delays[d] = symetric_filter_group_delay(filter.shape_ref(d),
						      symetries[d],
						      step,
						      extension);
}

static void
gen_convolver_output_lengths(dim_type dimension,
			     const index_type *input_shape,
			     const index_type *filter_shape,
			     index_type *output_shape,
			     int group_delay,
			     int step,
			     extension_type extension,
			     symetry_type symetry = non)
{
    dim_type d;

    for(d=0; d<dimension; d++) {
	switch(extension) {
	case Truncated:
	    output_shape[d] = (input_shape[d] 
			       - group_delay 
			       - filter_shape[d]
			       + 1);
	    break;
	    
	case Padded:
	    output_shape[d] = (input_shape[d]
			       + filter_shape[d]
			       - 1);
	    break;
	    
	case Symetric:
	    break;
	}
    }
}
	
	
/*
 * member function convolve
 *
 * This is the big one! It convolves an array with a kernel but that's not all.
 *
 * The function is designed to account for a range of signal extensions
 * which are implemented as extended array classes.  In addition, the array
 * may be resampled during the convolution operation (thus picking up on
 * the considerable efficiency improvements which this allows) for rate
 * conversion applications.
 *
 * The extensions etc. are defined in convolve.hh
 */ 

Num_Array*
_Num_Array::convolve_rows(const _Num_Array& kernel,
			 int step,
			 extension_type extension,
			 int aux_arg)	const
{
    index_type result_shape[MAX_DIMENSION];
    index_type input_index[MAX_DIMENSION];

    _Extended_Numarray	*_data = NULL;
    Num_Array		*result = NULL;

    int group_delay;
    index_type result_length;
    index_type length = kernel.size_ref();

    cbuff window(kernel.size_ref());		// data window
    T *kernel_data = new T[length];

    index_type i;
    
    switch(extension) {
    case Padded:	
	_data = new Padded__Numarray(*this, aux_arg);
	group_delay = length - 1;
	result_length = (index_type)ceil((double)(shape_ref(0)
						  + length
						  - 1) 
					 / (double)step);
	break;

    case Truncated:	
	_data = new Padded__Numarray(*this, 0);		
	group_delay = 0;
	result_length = (index_type)ceil((double)(shape_ref(0)
						  - length
						  + 1)
					 / (double)step);
	break;

    case Periodic:	
	_data = new Periodic__Numarray(*this);
	group_delay = aux_arg;
	result_length = (index_type)ceil((double)shape_ref(0) 
					 / (double)step);
	break;
    }

    memcpy(result_shape, _data->shape_ref(), 
	   dimension_ref() * sizeof(index_type));
    result_shape[0] = result_length;
    result = new Num_Array(dimension_ref(), result_shape);

    for(i=0; i<length; i++)	// copy (and reverse) the kernel data
	kernel_data[length-i-1] = kernel.elem_ref(&i);

    memset(input_index,  0, MAX_DIMENSION * sizeof(index_type));
    input_index[0] = -group_delay;
    for(i=0; i<length; i++, input_index[0]++)
	window.push(_data->elem_ref(input_index));

    unsigned long nout = result->size_ref();
	    
    /*
     * Convolve---elementwise multiplies and additions happen here
     */
    T* output_element = result->storage_iref();
    for(long j=0; j<nout; j++) {
	for(i=0, *output_element=0; i<length; i++)
	    *output_element += window.elem_ref(i) * kernel_data[i];
	/*
	 * Increment complex loop variables
	 */
	output_element++;
	if((j+1) % result->shape_ref(0)) {	// same row
	    for(int s=0; s<step; s++) {
		window.push(_data->elem_ref(input_index));
		input_index[0]++;
	    }
	}
	else {	// start new row
	    _data->inc_index(input_index);
	    input_index[0] = -group_delay;
	    for(i=0; i<length; i++, input_index[0]++)
		window.push(_data->elem_ref(input_index));
	}
    }

    return result;
}

Num_Array*
_Num_Array::convolve(const _Num_Array& kernel,
		     int step,
		     extension_type extension,
		     int aux_arg) const
{
    Num_Array	*result;

    if(kernel.dimension_ref() == 1) {	// assume separable convolution

	_Num_Array	*tmp_data = (_Num_Array *)this;
	dim_type d;
	
	for(d=0; d<dimension_ref(); d++) {
	    result = tmp_data->convolve_rows(kernel, step, extension, aux_arg);
	    if(tmp_data != this)
		delete tmp_data;
	    tmp_data = result;
	    tmp_data->rotate();
	}
    }
    else if(dimension_ref() == kernel.dimension_ref()) {// non separable kernel
    }	// non separable kernel
    return result;
}
		 
Num_Array*
_Num_Array::upsample_convolve_rows(const _Num_Array& kernel,
				   int step,
				   extension_type extension,
				   int aux_arg)	const
{
    index_type result_shape[MAX_DIMENSION];
    index_type input_index[MAX_DIMENSION];

    _Extended_Numarray	*_data = NULL;
    Num_Array		*result = NULL;

    int group_delay;
    index_type result_length;
    index_type length = kernel.size_ref();

    cbuff window(kernel.size_ref());		// data window
    T *kernel_data = new T[length];

    index_type i;
    
    switch(extension) {
    case Padded:	
	_data = new Padded__Numarray(*this, aux_arg);
	group_delay = length - 1;
	result_length = shape_ref(0) * step + length - 2;
	break;

    case Truncated:	
	_data = new Padded__Numarray(*this, 0);		
	group_delay = 0;
	result_length = shape_ref(0) * step - length + 1;
	break;

    case Periodic:
	_data = new Periodic__Numarray(*this);
	group_delay = aux_arg;
	result_length = shape_ref(0) * step;
	break;
    }

    memcpy(result_shape, _data->shape_ref(), 
	   dimension_ref() * sizeof(index_type));
    result_shape[0] = result_length;
    result = new Num_Array(dimension_ref(), result_shape);

    for(i=0; i<length; i++)	//copy and reverse the kernel data
	kernel_data[length-i-1] = kernel.elem_ref(&i);

    memset(input_index,  0, MAX_DIMENSION * sizeof(index_type));
    input_index[0] = (index_type)-floor((double)group_delay / (double)step);
    int step_count = -group_delay % step;
    for(i=0; i<length; i++, step_count=(step_count+1)%step) {
	if(step_count)
	    window.push(0);
	else {
	    window.push(_data->elem_ref(input_index));
	    input_index[0]++;
	}
    }

    unsigned long nout = result->size_ref();
	    
    /*
     * Convolve---elementwise multiplies and additions happen here
     */
    T *output_element = result->storage_iref();
    for(long j=0; j<nout; j++) {
	for(i=0, *output_element=0; i<length; i++)
	    *output_element += window.elem_ref(i) * kernel_data[i];
	/*
	 * Increment complex loop variables
	 */
	output_element++;
	if((j+1) % result->shape_ref(0)) {	// same row
	    if(step_count)
		window.push(0);
	    else {
		window.push(_data->elem_ref(input_index));
		input_index[0]++;
	    }
	    step_count = (step_count+1) % step;
	}
	else {	// start new row
	    _data->inc_index(input_index);
	    input_index[0] = (index_type)-floor((double)group_delay 
					       / (double)step);
	    step_count = -group_delay;
	    for(i=0; i<length; i++, step_count=(step_count+1)%step) {
		if(step_count)
		    window.push(0);
		else {
		    window.push(_data->elem_ref(input_index));
		    input_index[0]++;
		}
	    }
	}
    }
    
    return result;
}


Num_Array*
Num_Array::convolve_rows(const _Num_Array& kernel,
			 int step,
			 extension_type extension,
			 int aux_arg)	const
{
    index_type result_shape[MAX_DIMENSION];
    index_type input_index[MAX_DIMENSION];

    Extended_Numarray	*_data = NULL;
    Num_Array		*result = NULL;

    int group_delay;
    index_type result_length;
    index_type length = kernel.size_ref();

    cbuff window(kernel.size_ref());		// data window
    T *kernel_data = new T[length];

    index_type i;
    
    switch(extension) {
    case Padded:	
	_data = new Padded_Numarray(*this, aux_arg);
	group_delay = length - 1;
	result_length = (index_type)ceil((double)(shape_ref(0)
						  + length
						  - 1) 
					 / (double)step);
	break;

    case Truncated:	
	_data = new Padded_Numarray(*this, 0);		
	group_delay = 0;
	result_length = (index_type)ceil((double)(shape_ref(0)
						  - length
						  + 1)
					 / (double)step);
	break;

    case Periodic:	
	_data = new Periodic_Numarray(*this);
	group_delay = aux_arg;
	result_length = (index_type)ceil((double)shape_ref(0) 
					 / (double)step);
	break;
    }

    memcpy(result_shape, _data->shape_ref(), 
	   dimension_ref() * sizeof(index_type));
    result_shape[0] = result_length;
    result = new Num_Array(dimension_ref(), result_shape);

    for(i=0; i<length; i++)	// copy (and reverse) the kernel data
	kernel_data[length-i-1] = kernel.elem_ref(&i);

    memset(input_index,  0, MAX_DIMENSION * sizeof(index_type));
    input_index[0] = -group_delay;
    for(i=0; i<length; i++, input_index[0]++)
	window.push(_data->elem_ref(input_index));

    unsigned long nout = result->size_ref();
	    
    /*
     * Convolve---elementwise multiplies and additions happen here
     */
    T* output_element = result->storage_iref();
    for(long j=0; j<nout; j++) {
	for(i=0, *output_element=0; i<length; i++)
	    *output_element += window.elem_ref(i) * kernel_data[i];
	/*
	 * Increment complex loop variables
	 */
	output_element++;
	if((j+1) % result->shape_ref(0)) {	// same row
	    for(int s=0; s<step; s++) {
		window.push(_data->elem_ref(input_index));
		input_index[0]++;
	    }
	}
	else {	// start new row
	    _data->inc_index(input_index);
	    input_index[0] = -group_delay;
	    for(i=0; i<length; i++, input_index[0]++)
		window.push(_data->elem_ref(input_index));
	}
    }

    return result;
}

Num_Array*
Num_Array::convolve(const _Num_Array& kernel,
		    int step,
		    extension_type extension,
		    int aux_arg) const
{
    Num_Array	*result;

    if(kernel.dimension_ref() == 1) {	// assume separable convolution

	Num_Array	*tmp_data = (Num_Array *)this;
	dim_type d;
	
	for(d=0; d<dimension_ref(); d++) {
	    result = tmp_data->convolve_rows(kernel, step, extension, aux_arg);
	    if(tmp_data != this)
		delete tmp_data;
	    tmp_data = result;
	    tmp_data->rotate();
	}
    }
    else if(dimension_ref() == kernel.dimension_ref()) {// non separable kernel
    }	// non separable kernel
    return result;
}
		 
Num_Array*
Num_Array::upsample_convolve_rows(const _Num_Array& kernel,
				  int step,
				  extension_type extension,
				  int aux_arg)	const
{
    index_type result_shape[MAX_DIMENSION];
    index_type input_index[MAX_DIMENSION];

    Extended_Numarray	*_data = NULL;
    Num_Array		*result = NULL;

    int group_delay;
    index_type result_length;
    index_type length = kernel.size_ref();

    cbuff window(kernel.size_ref());		// data window
    T *kernel_data = new T[length];

    index_type i;
    
    switch(extension) {
    case Padded:	
	_data = new Padded_Numarray(*this, aux_arg);
	group_delay = length - 1;
	result_length = shape_ref(0) * step + length - 2;
	break;

    case Truncated:	
	_data = new Padded_Numarray(*this, 0);		
	group_delay = 0;
	result_length = shape_ref(0) * step - length + 1;
	break;

    case Periodic:
	_data = new Periodic_Numarray(*this);
	group_delay = aux_arg;
	result_length = shape_ref(0) * step;
	break;
    }

    memcpy(result_shape, _data->shape_ref(), 
	   dimension_ref() * sizeof(index_type));
    result_shape[0] = result_length;
    result = new Num_Array(dimension_ref(), result_shape);

    for(i=0; i<length; i++)	//copy and reverse the kernel data
	kernel_data[length-i-1] = kernel.elem_ref(&i);

    memset(input_index,  0, MAX_DIMENSION * sizeof(index_type));
    input_index[0] = (index_type)-floor((double)group_delay / (double)step);
    int step_count = -group_delay % step;
    for(i=0; i<length; i++, step_count=(step_count+1)%step) {
	if(step_count)
	    window.push(0);
	else {
	    window.push(_data->elem_ref(input_index));
	    input_index[0]++;
	}
    }

    unsigned long nout = result->size_ref();
	    
    /*
     * Convolve---elementwise multiplies and additions happen here
     */
    T* output_element = result->storage_iref();
    for(long j=0; j<nout; j++) {
	for(i=0, *output_element=0; i<length; i++)
	    *output_element += window.elem_ref(i) * kernel_data[i];
	/*
	 * Increment complex loop variables
	 */
	output_element++;
	if((j+1) % result->shape_ref(0)) {	// same row
	    if(step_count)
		window.push(0);
	    else {
		window.push(_data->elem_ref(input_index));
		input_index[0]++;
	    }
	    step_count = (step_count+1) % step;
	}
	else {	// start new row
	    _data->inc_index(input_index);
	    input_index[0] = (index_type)-floor((double)group_delay 
					       / (double)step);
	    step_count = -group_delay;
	    for(i=0; i<length; i++, step_count=(step_count+1)%step) {
		if(step_count)
		    window.push(0);
		else {
		    window.push(_data->elem_ref(input_index));
		    input_index[0]++;
		}
	    }
	}
    }
    
    return result;
}

