/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

/*
 * CONVOLVE.HH
 *
 * Function templates for convolution of 1D and 2D data sets
 * (Stupid C++ forces you to put your code in header files if you need 
 * genericity...)
 */

#ifndef __CONVOLVE__
#define __CONVOLVE__

#include <math.h>

enum extension_type	{Truncated, Padded, Periodic, Symetric, Arithmetic};
enum offset_type	{Zero, Half, One};
enum symetry_type	{non = 0,
			 wsw = 1, 
			 wsh = 2, 
			 hsw = 3, 
			 hsh = 4, 
			 waw = 5, 
			 wah = 6, 
			 haw = 7, 
			 hah = 8};


/*
 * CONVOLVE_2DNS_INSIDE(...)
 *
 * Truncated 2D non-separable convolution function.
 *
 * Allows fairly arbitrary indexing into data and result arrays by using
 * caller specified x(y) step and x(y) offset.  Decimated convolution may
 * be performed directly by setting the data x(y) step values to > 1.
 *
 * NB. Initialisation of the result array is NOT performed by this function.
 * The result is simply added to whatever is already present in the result
 * array.  This allows for increased efficiency where the result of a 
 * convolution is to be added to another data array (eg. unsharp masking).
 *
 * NBB. result is assumed to point to the first non-border location in the
 * result array.
 */

enum Edge_type { PERIODIC };
typedef unsigned int uint;

template<class KT,class DT,class RT>
void
convolve_2dns_inside(const KT *kernel, uint kx, uint ky,
		     const DT *data, uint dx, uint dy, 
		     uint dxoffset, uint dyoffset, uint dxstep,uint dystep,
		     RT *result, uint rx, uint ry, uint rxstep, uint rystep)
{
  uint x, y, i, j;
  const KT *k;
  const DT *d, *dr;	/* time savers into the data array		*/
  const DT *b, *br;	/* time savers into blocks in the data array	*/
  RT *r, *rr;		/* time savers into the result array		*/

  rystep*=rx;
  rx = (dx-dxoffset-kx+1) / dxstep;
  ry = (dy-dyoffset-ky+1) / dystep;
  dystep*=dx;

  for(y=0, dr=&data[dxoffset+dx*dyoffset], rr=result;  
      y<ry;  
      y++, dr+=dystep, rr+=rystep)
    for(x=0, d=dr, r=rr;   x<rx;   x++, d+=dxstep, r+=rxstep)
      for(k=kernel, j=0, br=d;   j<ky;   j++, br+=dy)
	for(i=0, b=br;   i<kx;   k++, i++, b++)
	  (*r) += (RT)((*k) * (*b));
}


/*
 * Some macros for handling boundaries for the convolution operation
 */
//
// We need an arithmetically sensible modulo operator.  As it is to be
// used in calculating indices into arrays it also needs to be efficient:
//
inline int Modulo(int n1, int n2) {
    int i = n1 % n2;
    return ( ((n1 < 0) != (n2 < 0) && (i != 0))  ? i+n2 : i );
}
template<class DT>
inline DT
PERIODIC_ACCESS2(DT *data, int dx, int dy, int x, int y)
{
  return data[Modulo(x,dx) + Modulo(y,dy) * dy];
}
// #define PERIODISE2(dx,dy,x,y)	((Modulo(x,dx) + Modulo(y,dy) * (dx)))
// #define PERIODIC_ACCESS2(data,dx,dy,x,y)   ((data)[PERIODISE2(dx,dy,x,y)])

/*
 * Convolve over the boundary set using periodic extension
 */ 
template<class KT,class DT,class RT>
void
convolve_2dns_boundary_periodic(const KT *kernel, uint kx, uint ky, 
				uint kxdelay, uint kydelay,
				const DT *data, uint dx, uint dy, 
				uint dxoffset, uint dyoffset,
				uint dxstep, uint dystep,
				RT *result, uint rx,
				uint rxstep, uint rystep
				)
{
  int xd, yd;
  uint xr, yr;
  uint i, j;
  const KT *k;
  RT *r;

  const uint datax   = (dx-kx-dxoffset+1);
  const uint datay   = (dy-ky-dyoffset+1);
  const uint insidex = (int)((double)datax / (double)dxstep) * rxstep;
  const uint insidey = (int)((double)datay / (double)dystep) * rystep;
  const uint topbw   = (int)((double)kydelay / (double)dystep) * rystep;
  const uint leftbw  = (int)(floor((double)kxdelay / (double)dxstep) * rxstep);
  const uint resultx = (int)ceil((double)dx / (double)dxstep) * rxstep;
  const uint resulty = (int)ceil((double)dy / (double)dystep) * rystep;

  /* 
   * Convolve top boundary:
   */
  for(yd=dyoffset-kydelay, yr=0; yr<topbw; yd+=dystep, yr+=rystep)
    for(xd=dxoffset-kxdelay, xr=0; xr<resultx; xr+=rxstep, xd+=dxstep)
      for(r=&result[xr+yr*rx], j=0, k=kernel; j<ky; j++)
	for(i=0; i<kx; i++, k++)
	  (*r)+=(RT)((*k)*PERIODIC_ACCESS2(data,dx,dy,xd+i,yd+j));

  /*
   * Then sides:
   */
  for(; yr<topbw+insidey; yd+=dystep, yr+=rystep) {
    /*
     * left border
     */
    for(xd=dxoffset-kxdelay, xr=0; xr<leftbw; xd+=dxstep, xr+=rxstep)
      for(r=&result[xr+yr*rx], j=0, k=kernel; j<ky; j++)
	for(i=0; i<kx; i++, k++)
	  (*r)+=(RT)((*k)*PERIODIC_ACCESS2(data,dx,dy,xd+i,yd+j));
    /*
     * right border (apologies for the lengthy initialisation!)
     */
    xd=(int)(ceil((float)(leftbw+insidex)/(float)rxstep)
	     * dxstep - kxdelay + dxoffset);
//    for(xd=(leftbw+insidex)/rxstep*dxstep-kxdelay+dxoffset,xr=leftbw+insidex;
    for(xr=leftbw+insidex; 
	xr<resultx; 
	xd+=dxstep,xr+=rxstep)
      for(r=&result[xr+yr*rx], j=0, k=kernel; j<ky; j++)
	for(i=0; i<kx; i++, k++)
	  (*r)+=(RT)((*k)*PERIODIC_ACCESS2(data,dx,dy,xd+i,yd+j));
  }

  /*
   * Then the bottom:
   */
  for(; yr<resulty; yd+=dystep, yr+=rystep)
    for(xd=dxoffset-kxdelay, xr=0; xr<resultx; xd+=dxstep, xr+=rxstep)
      for(r=&result[xr+yr*rx], j=0, k=kernel; j<ky; j++)
	for(i=0; i<kx; i++, k++)
	  (*r)+=(RT)((*k)*PERIODIC_ACCESS2(data,dx,dy,xd+i,yd+j));
}


/*
 * Combining the above gives the following template for 2d decimated periodic 
 * convolution.  It has been tested for 
 *
 * Note that this is given partly as a guide as it may often be desirable to
 * use the extra flexibility afforded by the lower level routines defined above
 *
 * NB. The convolve functions only work correctly for delay < kernel-size.
 */
template<class DT, class RT, class KT>
void
cds_2d_periodic(const KT *kernel, uint kx, uint ky,
		const DT *data, uint dx, uint dy,
		RT *result, uint rx, uint ry,
		uint xdelay=0, uint ydelay=0, 
		uint dxoffset=0, uint dyoffset=0,
		uint xrate=1,  uint yrate=1)
{
  /*
   * First determine the first data point involved in the convolution.
   * If the delay is a multiple of the rate then the first data sample
   * will be kept after downsampling otherwise, the start is a simple
   * function of rate and delay
   */
  dxoffset += (xdelay%xrate ? xrate-xdelay%xrate : 0);
  dyoffset += (ydelay%yrate ? yrate-ydelay%yrate : 0);
  /*
   * Determine the location of the "inside" convolution in the full result.
   */
  const int rxoffset = (int)ceil((double)xdelay / (double)xrate);
  const int ryoffset = (int)ceil((double)ydelay / (double)yrate);

  convolve_2dns_inside(kernel, kx, ky,
		       data, dx, dy, dxoffset, dyoffset, xrate, yrate,
		       &result[ryoffset*rx + rxoffset], rx, ry, 1, 1);
  convolve_2dns_boundary_periodic(kernel, kx, ky, xdelay, ydelay,
				  data, dx, dy, 
				  dxoffset, dyoffset, xrate, yrate,
				  result, rx, //ry, 
				  1, 1);
}

/*
 * The following demonstrates upsampled convolution implemented by interleaving
 * the results from convolutions with decimated versions of the kernel
 */
template<class DT, class RT, class KT>
void
cus_2d_periodic(const KT *kernel, uint kx, uint ky, 
		const DT *data,   uint dx, uint dy,
		RT *result, uint rx, uint ry,
		unsigned int xdelay=0, unsigned int ydelay=0, 
		unsigned int xrate=1,  unsigned int yrate=1)
{
  KT *dk, *deckern = (KT*)malloc(kx*ky*sizeof(KT));
  uint xk, yk, kxoffset, kyoffset, dxoffset, dyoffset, rxoffset, ryoffset;
  uint dkx, dky;

  /*
   * For each subkernel...
   */
  for(yk=0, kyoffset=ydelay%yrate, dyoffset=0, ryoffset=yrate*ydelay; 
      yk<2; 
      yk++, 
	kyoffset=(kyoffset+1)%yrate,
	dyoffset=1,
	ryoffset++)
    for(xk=0, kxoffset=xdelay%xrate, dxoffset=0, rxoffset=xrate*xdelay; 
	xk<2; 
	xk++,
	  kxoffset=(kxoffset+1)%xrate,
	  dxoffset=1,
	  rxoffset++) {
      /*
       * Extract subkernel coefficients
       */
      for(dky=kyoffset, dk=deckern; dky<ky; dky+=yrate)
	for(dkx=kxoffset; dkx<kx; dkx+=xrate)
	  *dk++ = kernel[dkx+dky*kx];
      dkx/=xrate;
      dky/=yrate;
      /*
       * And perform convolution
       */
      convolve_2dns_inside(deckern, dkx, dky,
			   data, dx, dy, dxoffset, dyoffset, 1, 1,
			   &result[ryoffset*rx+rxoffset], rx, ry, xrate,yrate);
      /*
       * Currently boundary handling is buggy
       */
      convolve_2dns_boundary_periodic(deckern, dkx, dky, xdelay, ydelay,
       				      data, dx, dy, 0, 0, 1, 1,
       				      &result[yk*rx+xk], rx, //ry, 
       				      xrate, yrate);
    }
}


#endif	//__CONVOLVE__
