/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

/*
 * Implementation of multiply functions for ND numeric arrays
 */

#include "Num_Array.hh"
#include "multiply.hh"
#include <values.h>
#include <math.h>

/*
 * matrix_multiply:  As the name suggests.  Each argument must have dimension
 *	of '2'; vectors (row or column) are indicated by one (second or first)
 *	element of the shape being '1'.
 */
void
matrix_multiply(const _Num_Array *A, const _Num_Array *B, _Num_Array *C)
{
    index_type ro,co, r, c, e;
    T t;
    index_type inda[2], indb[2], indc[2];

    co = B->shape_ref(0);	// number of columns in B
    ro = A->shape_ref(1);	// number of rows in A

    for(r=0; r<ro; r++) {	// for each row of output
	for(c=0; c<co; c++) {		//  for each col of output
	    for(e=0, t=0; e<A->shape_ref(0); e++) {//   for each col of A
		inda[0] = e; inda[1] = r;
		indb[0] = c; indb[1] = e;
		t += A->elem_ref(inda) * B->elem_ref(indb);
	    }
	    indc[0] = c; indc[1] = r;
	    C->elem_set(indc, t);
	}
    }
}

/*
 * scalar_multiply:  Returns the sum of the elementwise products of two
 * arrays.  This is the 'elementwise' extension of vector multiplication.
 * NB. if both args are vectors he same result can be obtained using matrix 
 * multiply---the result will be a single element array however.
 */
double scalar_multiply(const _Num_Array *A, const _Num_Array *B)
{
    unsigned long size;
    index_type ind[MAX_DIMENSION];
    double C;

    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));

    for(C=0, size=A->size_ref(); size>0; size--) {
	C += A->elem_ref(ind) * B->elem_ref(ind);
	A->inc_index(ind);
    }

    return C;
}

double scalar_multiply(const Num_Array *A, const Num_Array *B)
{
  double C = 0;
  
  if(A->convension_ref() == B->convension_ref()) {
    T *_A = A->storage_iref();
    T *_B = B->storage_iref();
    for(unsigned long size = A->size_ref(); size>0; size--, _A++, _B++)
      C += *_A * *_B;
  }
  else
    C = scalar_multiply((_Num_Array*)A, (_Num_Array*)B);
  
  return C;
}

double Lp_norm(const _Num_Array *A, int p)
{
  double C = 0;
  unsigned long size = A->size_ref();
  index_type index[MAX_DIMENSION];
  memset(index, 0, MAX_DIMENSION*sizeof(index_type));
  T t;

  switch(p) {

  case 1:
    for(; size>0; size--, A->inc_index(index)) {
      t =  A->elem_ref(index);
      C += (t>0 ? t : -t);
    }
    break;

  case 2:
    C = sqrt(scalar_multiply(A, A));
    break;

  case MAXINT:	// L_infinity norm
    for(C=A->elem_ref(index); size>0; size--, A->inc_index(index)) {
      t =  A->elem_ref(index);
      C = (t>0 ? (t>C ? t : C) : (-t>C ? -t : C));
    }
    break;
    
  default:
    for(; size>0; size--) {
      t =  A->elem_ref(index);
      C += (t>0 ? pow((double)t,(double)p) : pow((double)-t,(double)p));
    }
    C = pow(C, 1/(double)p);
  }

  return C;
}

