/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 * This file is a dirivitive work of other copyrighted software.  Please
 * see the file COPYRIGHTS for more information on these.
 */

/*
 * tkImgFmtNarray.c --
 *
 *	A photo image file handler for PPM (Portable PixMap) files.
 *
 * Author: Andrew Dorrell
 *	   School of Electrical Engineering
 *	   University of Technology, Sydney.
 */

#include "tkInt.h"
#include "tkPort.h"

/*
 * The maximum amount of memory to allocate for data read from the
 * file.  If we need more than this, we do it in pieces.
 */

#define MAX_MEMORY	10000		/* don't allocate > 10KB */

/*
 * The data types which may (eventually) be present in the data stream
 */
#define UNSUPPORTED	0
#define BYTE		1
#define SHORT		2
#define INT		3
#define FLOAT		5
#define DOUBLE		6

/*
 * The format record for the narray file (data) format:
 */

/* static int	FileMatchNarray   _ANSI_ARGS_((FILE *f, char *fileName, */
/* 					       char *formatString,  */
/* 					       int *widthPtr, */
/* 					       int *heightPtr)); */
static int	StringMatchNarray _ANSI_ARGS_((char *string, 
					       char *formatString, 
					       int *widthPtr, 
					       int *heightPtr));
/* static int	FileReadNarray    _ANSI_ARGS_((Tcl_Interp *interp, */
/* 					       FILE *f, char *fileName,  */
/* 					       char *formatString, */
/* 					       Tk_PhotoHandle imageHandle,  */
/* 					       int destX, int destY, */
/* 					       int width, int height,  */
/* 					       int srcX, int srcY)); */
static int	StringReadNarray  _ANSI_ARGS_((Tcl_Interp *interp,
					       char *string, 
					       char *formatString, 
					       Tk_PhotoHandle imageHandle,
					       int destX, int destY, 
					       int width, int height, 
					       int srcX, int srcY));
/* static int	FileWriteNarray   _ANSI_ARGS_((Tcl_Interp *interp, */
/* 					       char *fileName,  */
/* 					       char *formatString, */
/* 					       Tk_PhotoImageBlock *blockPtr)); */
/* static int	StringWriteNarray _ANSI_ARGS_((Tcl_Interp *interp, */
/* 					       Tcl_DString *dataPtr,  */
/* 					       char *formatString, */
/* 					       Tk_PhotoImageBlock *blockPtr)); */

Tk_PhotoImageFormat tkImgFmtNarray = {
    "NARRAY",					/* name */
    NULL,		/* FileMatchNarray,	/* fileMatchProc */
    StringMatchNarray,				/* stringMatchProc */
    NULL,		/* FileReadNarray,	/* fileReadProc */
    StringReadNarray,				/* stringReadProc */
    NULL,		/* FileWriteNarray,	/* fileWriteProc */
    NULL		/* StringWriteNarray,	/* stringWriteProc */
};

/*
 * Prototypes for local procedures defined in this file:
 */
static int
ReadNarrayStringHeader _ANSI_ARGS_((char *inString, char **outString, 
				    int  *widthPtr, int  *heightPtr, 
				    int  *depthPtr, int  *precPtr));


/*
 *----------------------------------------------------------------------
 *
 * StringMatchNarray --
 *
 *	This procedure is invoked by the photo image type to see if
 *	a string contains image data in Narray format.
 *
 * Results:
 *	The return value is >0 if the first characters in file "f" look
 *	like narray data, and 0 otherwise.
 *
 * Side effects:
 *	The access position in f may change.
 *
 *----------------------------------------------------------------------
 */

static int
StringMatchNarray(string, formatString, widthPtr, heightPtr)
    char *string;		/* The image data as a string. */
    char *formatString;		/* User-specified format string, or NULL. */
    int  *widthPtr, *heightPtr;	/* The dimensions of the image are
				 * returned here if the file is a valid
				 * raw narray file. */
{
    char *strPtr;
    int  depth;
    int  prec;

    return ReadNarrayStringHeader(string, &strPtr, 
				  widthPtr, heightPtr, &depth,
				  &prec);
}

/*
 *----------------------------------------------------------------------
 *
 * StringReadNarray --
 *
 *	This procedure is called by the photo image type to read
 *	narray format data from a file and write it into a given
 *	photo image.  NB The data part of the narray string is the
 *	address of the C data in memory.
 *
 * Results:
 *	A standard TCL completion code.  If TCL_ERROR is returned
 *	then an error message is left in interp->result.
 *
 * Side effects:
 *	The access position in file f is changed, and new data is
 *	added to the image given by imageHandle.
 *
 *----------------------------------------------------------------------
 */

static int
StringReadNarray(interp, string, formatString, imageHandle, destX, destY,
		 width, height, srcX, srcY)
    Tcl_Interp *interp;		/* Interpreter to use for reporting errors. */
    char *string;		/* String containing image data. */
    char *formatString;		/* User-specified format string, or NULL. */
    Tk_PhotoHandle imageHandle;	/* The photo image to write into. */
    int destX, destY;		/* Coordinates of top-left pixel in
				 * photo image to be written to. */
    int width, height;		/* Dimensions of block of photo image to
				 * be written to. */
    int srcX, srcY;		/* Coordinates of top-left pixel to be used
				 * in image being read. */
{
    char *strPtr = 0;
    int  dataWidth, dataHeight, dataDepth;
    int	g_offset, b_offset;
    int  prec;
    int  interleave;

    float *fdataPtr = 0;

    int nLines, nBytes, h, i;
    unsigned char *pixelPtr = 0;
    Tk_PhotoImageBlock block;

    if(!ReadNarrayStringHeader(string, &strPtr, 
			       &dataWidth, &dataHeight, &dataDepth,
			       &prec)) {
	Tcl_AppendResult(interp, "couldn't read narray headrer in string \"",
			 string, "\"", NULL);
	return TCL_ERROR;
    }

    if (prec == UNSUPPORTED) {
	Tcl_AppendResult(interp, "unsupported data type in string \"",
		string, "\"", NULL);
	return TCL_ERROR;
    }

    if ((srcX + width) > dataWidth) {
	width = dataWidth - srcX;
    }
    if ((srcY + height) > dataHeight) {
	height = dataHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
	|| (srcX >= dataWidth) || (srcY >= dataHeight)) {
	return TCL_OK;
    }

    if (dataDepth == 1) {	/* Assume greyscale	*/
        block.pixelSize = 1;
        block.offset[0] = 0;
	block.offset[1] = 0;
	block.offset[2] = 0;
    }
    else if(dataWidth==3 || dataHeight==3 || dataDepth==3) { /* Assume RGB */
        block.pixelSize = 3;
        block.offset[0] = 0;
	block.offset[1] = 1;
	block.offset[2] = 2;
    }

    if(dataHeight==3)		interleave = 1;
    else if(dataDepth==3)	interleave = 2;
    else			interleave = 0;

    block.width = width;
    block.pitch = block.pixelSize * dataWidth;
    g_offset = dataHeight * dataWidth;
    b_offset = 2 * g_offset;

    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

    if(prec == FLOAT) {
	sscanf(strPtr, "%x", &fdataPtr);
	if (srcY > 0)
	    fdataPtr += srcY * block.pitch * sizeof(float);

	nLines = (MAX_MEMORY + block.pitch - 1) / block.pitch;
	if (nLines > height) {
	    nLines = height;
	}
	if (nLines <= 0) {
	    nLines = 1;
	}
	nBytes = nLines * block.pitch;
	pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
	block.pixelPtr = pixelPtr + srcX * block.pixelSize;

	for (h = height; h > 0; h -= nLines) {
	    if (nLines > h) {
		nLines = h;
		nBytes = nLines * block.pitch;
	    }
	    switch(interleave) {
	    case 0:
		for(i=0; i<nBytes; i++) {
		    pixelPtr[i] = (unsigned char)(*fdataPtr < 0.0 ? 0 :
						  *fdataPtr > 255.0 ? 255 : 
						  *fdataPtr);
		    fdataPtr++;
		}
		break;
	    case 1:
	    case 2:
		for(i=0; i<nBytes; i+=3) {
		    pixelPtr[i+0] = *fdataPtr;
		    pixelPtr[i+1] = fdataPtr[g_offset];
		    pixelPtr[i+2] = fdataPtr[b_offset];
		    fdataPtr++;
		}
		break;
	    }
	    block.height = nLines;
	    Tk_PhotoPutBlock(imageHandle, &block, destX, destY, width, nLines);
	    destY += nLines;
	}
    }

    ckfree((char *) pixelPtr);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ReadNarrayStringHeader --
 *
 *	This procedure reads the narray header from the beginning of a
 *	narray string and returns information from the header.
 *
 * Results:
 *	The return value is non-zero if inString appears to start with
 *	a valid narray header, and 0 otherwise.  If the header is valid,
 *	then *widthPtr, *heightPtr and depthPtr  are modified to hold the
 *	dimensions of the image and *precPtr is modified to
 *	hold the precision of the stored data (FLOAT is currently the only
 *	inplemented precision).  strPtr is used to point to the next
 *	character in inString past the header info --- ie. the start of the
 *	data.
 *
 * Side effects:
 *	None
 *
 *----------------------------------------------------------------------
 */

static int
ReadNarrayStringHeader(inString, strPtr,
		       widthPtr, heightPtr, depthPtr, 
		       precPtr)
    char *inString;		/* String representation of narray data	*/
    char **strPtr;		/* String adjusted to just past the header */
    int *widthPtr, *heightPtr,	/* The dimensions of the image are */
        *depthPtr;		/* returned here. */
    int *precPtr;		/* The precision of the data -- will be FLOAT
				   presently as no other types are available */
{
    /*
     * Read 4 space-separated fields from the file, ignoring
     * comments (any line that starts with "#").
     */
    int  is_narray = 1;
    int  index;
    char *magic_number = "#[narray";
    int  magic_number_length = 8;
    char precStr[32];

    for(index=0; index< magic_number_length; index++)
	is_narray &= (magic_number[index] == inString[index]);
    
    if(is_narray) {
	*strPtr = strchr(inString, '(');
	switch(sscanf(*strPtr, "(%i %i %i) <", 
		      widthPtr, heightPtr, depthPtr)) 
	    {
	    case 2:	*depthPtr = 1;	/* treat as greyscale	*/
	    case 3:	break;		/* treat as colour	*/
	    default:	is_narray = 0;	/* Is really but it is not an image */
	    }
	*strPtr = strchr(*strPtr, '<');
	sscanf(*strPtr, "<%s", precStr);
	if(!strcmp(precStr, "FLOAT")) {
	    *strPtr += 6;
	    *precPtr = FLOAT;
	}
	else
	    *precPtr = UNSUPPORTED;
    }
    
    return is_narray;
}


/*
 *----------------------------------------------------------------------
 *
 * Narray_Init --
 *
 *	This procedure is typically invoked by Tcl_AppInit procedures
 *	to perform additional initialization.
 *
 * Results:
 *	Returns a standard Tcl completion code and sets interp->result
 *	if there is an error.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
Narray_Init(interp)
    Tcl_Interp *interp;		/* Interpreter to initialize. */
{
    Tk_CreatePhotoImageFormat(&tkImgFmtNarray);

    return TCL_OK;
}


